import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    add_pre_post_process,
    ZeroInitGain,
)
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.coco import eval_func, anchors_func

PRETRAINED_DICT = {
    "yolov8_l": {"coco": "db0fcde5d2811b33b7f5f0f400d76911"},
    "yolov8_m": {"coco": "cb8c25148bb17485776ade4cf80cc6f6"},
    "yolov8_n": {"coco": "4cb83c7e452cdcd440b75546df0b211e"},
    "yolov8_s": {"coco": "4e1ac133e2a8831845172d8491c2747a"},
    "yolov8_x": {"coco": "2be28e650bf299aeea7ee26ab765a23e"},
    "yolov8_x6": {"coco": "f51ed830ccf5efae7dc56f2ce5e20890"},
    "yolov8_l_cls": {"imagenet": "071f41125034dd15401f6c6925fc1e6f"},
    "yolov8_m_cls": {"imagenet": "35ef50aa07ff232afa08f321447e354d"},
    "yolov8_n_cls": {"imagenet": "b1cfac787589689c0f2abde6893ec140"},
    "yolov8_s_cls": {"imagenet": "2caa57e8cf67b39921c35f89cea5061c"},
    "yolov8_x_cls": {"imagenet": "2d4b8b996c24f5fde903678ee8b7cf20"},
}


""" Yolov8Backbone """
BATCH_NORM_EPSILON = 1e-3
BATCH_NORM_MOMENTUM = 0.97


def conv_bn(inputs, output_channel, kernel_size=1, strides=1, use_bias=False, activation="swish", name=""):
    # print(f">>>> {inputs.shape = }, {output_channel = }, {kernel_size = }, {strides = }")
    nn = conv2d_no_bias(inputs, output_channel, kernel_size, strides, use_bias=use_bias, padding="same", name=name)
    return batchnorm_with_activation(nn, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name)


def reparam_conv_bn(inputs, output_channel, kernel_size=3, strides=1, use_bias=False, use_identity=True, activation="swish", name=""):
    branch_3x3 = conv2d_no_bias(inputs, output_channel, 3, strides, use_bias=False, padding="same", name=name + "REPARAM_k3_")
    branch_3x3 = batchnorm_with_activation(branch_3x3, activation=None, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name + "REPARAM_k3_")

    branch_1x1 = conv2d_no_bias(inputs, output_channel, 1, strides, use_bias=use_bias, padding="valid", name=name + "REPARAM_k1_")
    # branch_1x1 = batchnorm_with_activation(branch_1x1, activation=None, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name + "REPARAM_k1_")

    # out = (branch_3x3 + branch_1x1 + inputs) if use_identity else (branch_3x3 + branch_1x1)
    out = layers.Add(name=name + "REPARAM_out")([branch_3x3, branch_1x1, inputs] if use_identity else [branch_3x3, branch_1x1])
    return batchnorm_with_activation(out, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name)


def csp_with_2_conv(
    inputs,
    channels=-1,
    depth=2,
    shortcut=True,
    expansion=0.5,
    parallel_mode=True,
    use_bias=False,
    use_alpha=False,
    use_reparam_conv=False,
    activation="swish",
    name="",
):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    channels = channels if channels > 0 else inputs.shape[channel_axis]
    hidden_channels = int(channels * expansion)

    # short = conv_bn(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "short_")  # For YOLO_NAS
    # deep = conv_bn(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "deep_")  # For YOLO_NAS
    pre = conv_bn(inputs, hidden_channels * 2, kernel_size=1, activation=activation, name=name + "pre_")
    if parallel_mode:
        short, deep = functional.split(pre, 2, axis=channel_axis)
    else:  # parallel_mode=False for YOLOV8_X6 `path_aggregation_fpn` C2 module, deep branch first
        deep, short = functional.split(pre, 2, axis=channel_axis)

    out = [short, deep]
    for id in range(depth):
        cur_short = ZeroInitGain(name=name + "short_{}_alpha".format(id))(out[-1]) if use_alpha else out[-1]
        cur_name = name + "pre_{}_".format(id)
        if use_reparam_conv:
            deep = reparam_conv_bn(deep, hidden_channels, kernel_size=3, use_bias=use_bias, activation=activation, name=cur_name + "1_")
            deep = reparam_conv_bn(deep, hidden_channels, kernel_size=3, use_bias=use_bias, activation=activation, name=cur_name + "2_")
        else:
            deep = conv_bn(deep, hidden_channels, kernel_size=3, use_bias=use_bias, activation=activation, name=cur_name + "1_")
            deep = conv_bn(deep, hidden_channels, kernel_size=3, use_bias=use_bias, activation=activation, name=cur_name + "2_")

        deep = (cur_short + deep) if shortcut else deep
        out.append(deep)
    # parallel_mode=False for YOLOV8_X6 `path_aggregation_fpn` C2 module, only concat `short` and the last `deep` one.
    out = functional.concat(out, axis=channel_axis) if parallel_mode else functional.concat([deep, short], axis=channel_axis)
    # out = functional.concat([*out[1:], out[0]], axis=channel_axis) if parallel_mode else functional.concat([deep, short], axis=channel_axis)  # For YOLO_NAS
    out = conv_bn(out, channels, kernel_size=1, activation=activation, name=name + "output_")
    return out


def spatial_pyramid_pooling_fast(inputs, pool_size=5, activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channels = inputs.shape[channel_axis]
    hidden_channels = int(input_channels // 2)

    nn = conv_bn(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "pre_")
    pool_1 = layers.MaxPool2D(pool_size=pool_size, strides=1, padding="same")(nn)
    pool_2 = layers.MaxPool2D(pool_size=pool_size, strides=1, padding="same")(pool_1)
    pool_3 = layers.MaxPool2D(pool_size=pool_size, strides=1, padding="same")(pool_2)

    out = functional.concat([nn, pool_1, pool_2, pool_3], axis=channel_axis)
    out = conv_bn(out, input_channels, kernel_size=1, activation=activation, name=name + "output_")
    return out


def YOLOV8Backbone(
    channels=[32, 64, 128, 256],
    depthes=[1, 2, 2, 1],
    out_features=[-3, -2, -1],
    csp_expansions=0.5,
    csp_parallel_mode=True,
    use_alpha=False,
    use_bias=False,
    use_reparam_conv=False,
    input_shape=(640, 640, 3),
    activation="swish",
    num_classes=0,  # > 0 value for classification model
    dropout=0,  # for classification model
    classifier_activation="softmax",  # for classification model
    pretrained=None,  # for classification model
    model_name="yolov8_backbone",
    kwargs=None,  # Not using, recieving parameter
):
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape))
    is_classification_model = num_classes > 0

    global BATCH_NORM_EPSILON
    global BATCH_NORM_MOMENTUM
    if is_classification_model:
        BATCH_NORM_EPSILON = 1e-5
        BATCH_NORM_MOMENTUM = 0.9
    else:
        BATCH_NORM_EPSILON = 1e-3
        BATCH_NORM_MOMENTUM = 0.97

    """ Stem """
    # stem_width = stem_width if stem_width > 0 else channels[0]
    stem_width = channels[0]
    if use_reparam_conv:
        nn = reparam_conv_bn(inputs, stem_width // 2, kernel_size=3, strides=2, use_bias=use_bias, use_identity=False, activation=activation, name="stem_1_")
        nn = reparam_conv_bn(nn, stem_width, kernel_size=3, strides=2, use_bias=use_bias, use_identity=False, activation=activation, name="stem_2_")
    else:
        nn = conv_bn(inputs, stem_width // 2, kernel_size=3, strides=2, use_bias=use_bias, activation=activation, name="stem_1_")
        nn = conv_bn(nn, stem_width, kernel_size=3, strides=2, use_bias=use_bias, activation=activation, name="stem_2_")

    """ blocks """
    block_kwargs = dict(use_bias=use_bias, use_alpha=use_alpha, use_reparam_conv=use_reparam_conv, activation=activation)
    features = [nn]
    for stack_id, (channel, depth) in enumerate(zip(channels, depthes)):
        stack_name = "stack{}_".format(stack_id + 1)
        cur_name = stack_name + "downsample_"
        if stack_id >= 1:
            if use_reparam_conv:
                nn = reparam_conv_bn(nn, channel, kernel_size=3, strides=2, use_bias=use_bias, use_identity=False, activation=activation, name=cur_name)
            else:
                nn = conv_bn(nn, channel, kernel_size=3, strides=2, use_bias=use_bias, activation=activation, name=cur_name)
        csp_expansion = csp_expansions[stack_id] if isinstance(csp_expansions, (list, tuple)) else csp_expansions
        parallel_mode = csp_parallel_mode[stack_id] if isinstance(csp_parallel_mode, (list, tuple)) else csp_parallel_mode
        nn = csp_with_2_conv(nn, depth=depth, expansion=csp_expansion, parallel_mode=parallel_mode, **block_kwargs, name=stack_name + "c2f_")

        if not is_classification_model and stack_id == len(depthes) - 1:
            nn = spatial_pyramid_pooling_fast(nn, pool_size=5, activation=activation, name=stack_name + "spp_fast_")
        features.append(nn)

    if is_classification_model:
        nn = conv_bn(nn, 1280, kernel_size=1, strides=1, activation=activation, name="pre_")
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        outputs = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)
    else:
        outputs = [features[ii] for ii in out_features]
    model = models.Model(inputs, outputs, name=model_name)

    if is_classification_model:
        add_pre_post_process(model, rescale_mode="raw01")
        reload_model_weights(model, PRETRAINED_DICT, "yolov8", pretrained)
    return model


""" path aggregation fpn """


def path_aggregation_fpn(features, depth=3, parallel_mode=True, use_reparam_conv=False, activation="swish", name=""):
    # yolov8
    # 9: p5 1024 ---+----------------------+-> 21: out2 1024
    #               v [up 1024 -> concat]  ^ [down 512 -> concat]
    # 6: p4 512 --> 12: p4p5 512 --------> 18: out1 512
    #               v [up 512 -> concat]   ^ [down 256 -> concat]
    # 4: p3 256 --> 15: p3p4p5 256 --------+--> 15: out0 128
    # features: [p3, p4, p5]
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    upsamples = [features[-1]]
    p_name = "p{}_".format(len(features) + 2)
    # upsamples: [p5], features[:-1][::-1]: [p4, p3] -> [p5, p4p5, p3p4p5]
    for id, feature in enumerate(features[:-1][::-1]):
        cur_p_name = "p{}".format(len(features) + 1 - id)
        p_name = cur_p_name + p_name
        size = functional.shape(feature)[1:-1] if image_data_format() == "channels_last" else functional.shape(feature)[2:]
        nn = functional.resize(upsamples[-1], size, method="nearest")
        nn = functional.concat([nn, feature], axis=channel_axis)

        out_channel = feature.shape[channel_axis]
        nn = csp_with_2_conv(
            nn, out_channel, depth, shortcut=False, parallel_mode=parallel_mode, use_reparam_conv=use_reparam_conv, activation=activation, name=name + p_name
        )
        upsamples.append(nn)

    downsamples = [upsamples[-1]]
    # downsamples: [p3p4p5], upsamples[:-1][::-1]: [p4p5, p5] -> [p3p4p5, p3p4p5 + p4p5, p3p4p5 + p4p5 + p5]
    for id, ii in enumerate(upsamples[:-1][::-1]):
        cur_name = name + "c3n{}_".format(id + 3)
        nn = conv_bn(downsamples[-1], downsamples[-1].shape[channel_axis], kernel_size=3, strides=2, activation=activation, name=cur_name + "down_")
        nn = functional.concat([nn, ii], axis=channel_axis)

        out_channel = ii.shape[channel_axis]
        nn = csp_with_2_conv(
            nn, out_channel, depth=depth, shortcut=False, parallel_mode=parallel_mode, use_reparam_conv=False, activation=activation, name=cur_name
        )
        downsamples.append(nn)
    return downsamples


def yolov8_head(
    inputs,
    num_classes=80,
    regression_len=64,
    num_anchors=1,
    depth=2,
    hidden_channels=-1,
    use_object_scores=False,
    activation="swish",
    classifier_activation="sigmoid",
    name="",
):
    channel_axis = -1 if image_data_format() == "channels_last" else 1

    outputs = []
    if hidden_channels == -1:
        reg_channel = max(64, regression_len, inputs[0].shape[channel_axis] // 4)
        cls_channel = max(num_classes, inputs[0].shape[channel_axis])
        reg_channels, cls_channels = [reg_channel] * len(inputs), [cls_channel] * len(inputs)
    elif isinstance(hidden_channels, (list, tuple)):
        reg_channels, cls_channels = hidden_channels, hidden_channels

    for id, (feature, reg_channel, cls_channel) in enumerate(zip(inputs, reg_channels, cls_channels)):
        cur_name = name + "{}_".format(id + 1)

        reg_nn = feature
        for id in range(depth):
            reg_nn = conv_bn(reg_nn, reg_channel, 3, activation=activation, name=cur_name + "reg_{}_".format(id + 1))
        reg_out = conv2d_no_bias(reg_nn, regression_len * num_anchors, 1, use_bias=True, bias_initializer="ones", name=cur_name + "reg_3_")

        strides = 2 ** (id + 3)
        bias_init = initializers.constant(math.log(5 / num_classes / (640 / strides) ** 2))
        cls_nn = feature
        for id in range(depth):
            cls_nn = conv_bn(cls_nn, cls_channel, 3, activation=activation, name=cur_name + "cls_{}_".format(id + 1))
        cls_out = conv2d_no_bias(cls_nn, num_classes * num_anchors, 1, use_bias=True, bias_initializer=bias_init, name=cur_name + "cls_3_")
        if classifier_activation is not None:
            cls_out = activation_by_name(cls_out, classifier_activation, name=cur_name + "classifier_")

        # obj_preds, not using for yolov8
        if use_object_scores:
            bias_init = initializers.constant(-math.log((1 - 0.01) / 0.01))
            obj_out = conv2d_no_bias(reg_nn, 1 * num_anchors, kernel_size=1, use_bias=True, bias_initializer=bias_init, name=cur_name + "object_")
            obj_out = activation_by_name(obj_out, classifier_activation, name=cur_name + "object_out_")
            out = functional.concat([reg_out, cls_out, obj_out], axis=channel_axis)
        else:
            out = functional.concat([reg_out, cls_out], axis=channel_axis)
        out = out if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(out)
        out = layers.Reshape([-1, out.shape[-1]], name=cur_name + "output_reshape")(out)
        outputs.append(out)
    outputs = functional.concat(outputs, axis=1)
    return outputs


""" YOLOV8 models """


def YOLOV8(
    backbone=None,
    csp_channels=[32, 64, 128, 256],  # [YOLOV8Backbone parameters]
    csp_depthes=[1, 2, 2, 1],
    features_pick=[-3, -2, -1],  # [Detector parameters]
    regression_len=64,  # bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64
    use_reparam_conv=False,  # Use reparam_conv_bn instead of conv_bn block in all csp_blocks.
    paf_parallel_mode=True,  # paf_parallel_mode=False for YOLOV8_X6 `path_aggregation_fpn` module, only concat `short` and the last `deep` one.
    anchors_mode="yolov8",
    num_anchors="auto",  # "auto" means: anchors_mode=="anchor_free" / "yolov8" -> 1, anchors_mode=="yolor" -> 3, else 9
    use_object_scores="auto",  # "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False
    input_shape=(640, 640, 3),
    num_classes=80,
    activation="swish",
    classifier_activation="sigmoid",
    freeze_backbone=False,
    pretrained=None,
    model_name="yolov8",
    pyramid_levels_min=3,  # Init anchors for model prediction.
    anchor_scale="auto",  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
    rescale_mode="raw01",  # For decode predictions, raw01 means input value in range [0, 1].
    kwargs=None,  # Not using, recieving parameter
):
    if backbone is None:
        backbone = YOLOV8Backbone(
            csp_channels, csp_depthes, features_pick, use_reparam_conv=use_reparam_conv, input_shape=input_shape, activation=activation, model_name="backbone"
        )
        features = backbone.outputs
    else:
        if isinstance(features_pick[0], str):
            features = [backbone.get_layer(layer_name) for layer_name in features_pick]
        else:
            features = model_surgery.get_pyramide_feature_layers(backbone)
            features = [features[id] for id in features_pick]
        feature_names, features = model_surgery.align_pyramide_feature_output_by_image_data_format(features)
        print(">>>> features:", {ii: jj.shape for ii, jj in zip(feature_names, features)})

    backbone.trainable = False if freeze_backbone else True
    use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, num_anchors, anchor_scale)
    inputs = backbone.inputs[0]

    fpn_features = path_aggregation_fpn(features, csp_depthes[-1], paf_parallel_mode, use_reparam_conv=use_reparam_conv, activation=activation, name="pafpn_")

    header_kwargs = {"use_object_scores": use_object_scores, "activation": activation, "classifier_activation": classifier_activation}
    outputs = yolov8_head(fpn_features, num_classes, regression_len, num_anchors, **header_kwargs, name="head_")
    outputs = layers.Activation("linear", dtype="float32", name="outputs_fp32")(outputs)

    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "yolov8", pretrained)

    pyramid_levels = [pyramid_levels_min, pyramid_levels_min + len(features_pick) - 1]  # -> [3, 5]
    post_process = eval_func.DecodePredictions(
        backbone.input_shape[1:], pyramid_levels, anchors_mode, use_object_scores, anchor_scale, regression_len=regression_len
    )
    add_pre_post_process(model, rescale_mode=rescale_mode, post_process=post_process)
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


def switch_to_deploy(model):
    from keras_cv_attention_models.model_surgery.model_surgery import fuse_reparam_blocks, convert_to_fused_conv_bn_model

    new_model = convert_to_fused_conv_bn_model(fuse_reparam_blocks(convert_to_fused_conv_bn_model(model)))

    # Has to create a new one, or will `ValueError: Unable to create group (name already exists)` when saving
    post_process = eval_func.DecodePredictions(
        input_shape=model.decode_predictions.__input_shape__,
        pyramid_levels=model.decode_predictions.pyramid_levels,
        anchors_mode=model.decode_predictions.anchors_mode,
        use_object_scores=model.decode_predictions.use_object_scores,
        anchor_scale=model.decode_predictions.anchor_scale,
        regression_len=model.decode_predictions.regression_len,
    )
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=post_process)
    return new_model


@register_model
def YOLOV8_N(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    return YOLOV8(**locals(), model_name=kwargs.pop("model_name", "yolov8_n"), **kwargs)


@register_model
def YOLOV8_S(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_channels = [64, 128, 256, 512]
    return YOLOV8(**locals(), model_name=kwargs.pop("model_name", "yolov8_s"), **kwargs)


@register_model
def YOLOV8_M(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_channels = [96, 192, 384, 576]
    csp_depthes = [2, 4, 4, 2]
    return YOLOV8(**locals(), model_name=kwargs.pop("model_name", "yolov8_m"), **kwargs)


@register_model
def YOLOV8_L(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_channels = [128, 256, 512, 512]
    csp_depthes = [3, 6, 6, 3]
    return YOLOV8(**locals(), model_name=kwargs.pop("model_name", "yolov8_l"), **kwargs)


@register_model
def YOLOV8_X(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_channels = [160, 320, 640, 640]
    csp_depthes = [3, 6, 6, 3]
    return YOLOV8(**locals(), model_name=kwargs.pop("model_name", "yolov8_x"), **kwargs)


@register_model
def YOLOV8_X6(input_shape=(1280, 1280, 3), freeze_backbone=False, num_classes=80, backbone=None, classifier_activation="sigmoid", pretrained="coco", **kwargs):
    csp_channels = [160, 320, 640, 640, 640]
    csp_depthes = [3, 6, 6, 3, 3]
    features_pick = [-4, -3, -2, -1]
    paf_parallel_mode = False  # C2
    return YOLOV8(**locals(), model_name=kwargs.pop("model_name", "yolov8_x6"), **kwargs)


""" Classification models """


@register_model
def YOLOV8_N_CLS(input_shape=(640, 640, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return YOLOV8Backbone(**locals(), model_name=kwargs.pop("model_name", "yolov8_n_cls"), **kwargs)


@register_model
def YOLOV8_S_CLS(input_shape=(640, 640, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    channels = [64, 128, 256, 512]
    return YOLOV8Backbone(**locals(), model_name=kwargs.pop("model_name", "yolov8_s_cls"), **kwargs)


@register_model
def YOLOV8_M_CLS(input_shape=(640, 640, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    channels = [96, 192, 384, 768]
    depthes = [2, 4, 4, 2]
    return YOLOV8Backbone(**locals(), model_name=kwargs.pop("model_name", "yolov8_m_cls"), **kwargs)


@register_model
def YOLOV8_L_CLS(input_shape=(640, 640, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    channels = [128, 256, 512, 1024]
    depthes = [3, 6, 6, 3]
    return YOLOV8Backbone(**locals(), model_name=kwargs.pop("model_name", "yolov8_l_cls"), **kwargs)


@register_model
def YOLOV8_X_CLS(input_shape=(640, 640, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    channels = [160, 320, 640, 1280]
    depthes = [3, 6, 6, 3]
    return YOLOV8Backbone(**locals(), model_name=kwargs.pop("model_name", "yolov8_x_cls"), **kwargs)



import inspect
import torch
import numpy as np
from torch import nn
from functools import partial
from keras_cv_attention_models.pytorch_backend import initializers


""" Basic Layers """


def to_tuple(param, num_dims=2):
    return tuple(param if isinstance(param, (list, tuple)) else [param] * num_dims)


def get_perm(total_axis, from_axis, to_axis):
    # total_axis, from_axis, to_axis = 4, 1, 3 -> [0, 2, 3, 1]
    # total_axis, from_axis, to_axis = 4, 3, 1 -> [0, 3, 1, 2]
    from_axis = (total_axis + from_axis) if from_axis < 0 else from_axis
    to_axis = (total_axis + to_axis) if to_axis < 0 else to_axis
    aa = [ii for ii in range(total_axis) if ii != from_axis]
    aa.insert(to_axis, from_axis)
    return aa


def tf_same_pad(size, kernel_size, stride, dilation_rate=1):
    if size is None:
        return kernel_size - stride  # Regarding as size % stride == 0
    else:
        return max((np.math.ceil(size / stride) - 1) * stride + (kernel_size - 1) * dilation_rate + 1 - size, 0)


def compute_conv_output_size(input_shape, kernel_size, strides=1, padding="valid", dilation_rate=1):
    size = input_shape if len(input_shape) == 2 else input_shape[2:]
    kernel_size = kernel_size if isinstance(kernel_size, (list, tuple)) else [kernel_size] * len(size)
    dilation_rate = dilation_rate if isinstance(dilation_rate, (list, tuple)) else [dilation_rate] * len(size)
    strides = strides if isinstance(strides, (list, tuple)) else [strides] * len(size)
    dilated_filter_size = [kk + (kk - 1) * (dd - 1) for kk, dd in zip(kernel_size, dilation_rate)]
    if (isinstance(padding, str) and padding.lower() == "valid") or (isinstance(padding, (list, tuple)) and max(padding) == 0) or padding == 0:
        size = [None if ii is None else (ii - jj + 1) for ii, jj in zip(size, dilated_filter_size)]
    size = [None if ii is None else ((ii + jj - 1) // jj) for ii, jj in zip(size, strides)]
    return size


class Weight:
    def __init__(self, name, value):
        self.name, self.shape, self.__value__ = name, value.shape, value
        self.trainable = value.requires_grad if hasattr(value, "requires_grad") else True

    def __repr__(self):
        return "{}, shape={}".format(self.name, self.shape)

    @property
    def dtype(self):
        return self.__value__.dtype

    def value(self):
        return self.__value__

    def numpy(self):
        return self.__value__.detach().cpu().numpy()


class GraphNode:
    num_instances = 0  # Count instances

    @classmethod
    def __count__(cls):
        cls.num_instances += 1

    def __init__(self, shape, name=None):
        self.shape = shape
        self.name = "graphnode_{}".format(self.num_instances) if name is None else name
        self.pre_nodes, self.pre_node_names, self.next_nodes, self.next_node_names = [], [], [], []
        self.module = lambda xx: xx
        self.dtype = torch.get_default_dtype()
        self.__count__()

    def __repr__(self):
        # return ",".join(for kk, vv in zip())
        if len(self.pre_nodes) == 0:
            rr = "{}:\n  shape: {}\n  dtype: {}".format(self.name, self.shape, self.dtype)
        else:
            rr = "{}:\n  pre_nodes: {}".format(self.name, {ii.name: ii.shape for ii in self.pre_nodes})
        rr += "\n  next_nodes: {}".format({ii.name: ii.shape for ii in self.next_nodes})
        if hasattr(self, "layer") and hasattr(self.layer, "output_shape"):
            rr += "\n  out: {}".format(self.layer.output_shape)
        return rr

    def __getitem__(self, index_expr):
        # print(index_expr)
        # return Slice(index_expr)(self)
        index_expr = index_expr if isinstance(index_expr, (int, slice)) else tuple(index_expr)
        return Lambda(lambda inputs: inputs[index_expr])(self)

    def __add__(self, another):
        return Add()([self, another]) if isinstance(another, GraphNode) else Lambda(lambda xx: xx + another)(self)

    def __radd__(self, another):
        return self.__add__(another)

    def __sub__(self, another):
        return Subtract()([self, another]) if isinstance(another, GraphNode) else Lambda(lambda xx: xx - another)(self)

    def __rsub__(self, another):
        return self.__sub__(another)

    def __mul__(self, another):
        return Multiply()([self, another]) if isinstance(another, GraphNode) else Lambda(lambda xx: xx * another)(self)

    def __rmul__(self, another):
        return self.__mul__(another)

    def __truediv__(self, another):
        return Divide()([self, another]) if isinstance(another, GraphNode) else Lambda(lambda xx: xx / another)(self)

    def __rtruediv__(self, another):
        return self.__truediv__(another)

    # def __floordiv__(self, another):
    #     return Lambda(lambda xx: xx // another)(self)

    def __matmul__(self, another):
        # return Matmul()([self, another])
        return Lambda(lambda inputs: torch.matmul(inputs[0], inputs[1]))([self, another])

    def __pow__(self, exponent):
        # return Lambda(lambda inputs: torch.pow(inputs, exponent))(self)
        return Lambda(partial(torch.pow, exponent=exponent))(self)

    def set_shape(self, shape):
        self.shape = shape

    def set_pre_nodes(self, pre_nodes):
        pre_nodes = [ii for ii in pre_nodes if isinstance(ii, GraphNode)] if isinstance(pre_nodes, (list, tuple)) else [pre_nodes]
        self.pre_nodes += pre_nodes
        self.pre_node_names += [ii.name for ii in pre_nodes]

    def set_next_nodes(self, next_nodes):
        next_nodes = next_nodes if isinstance(next_nodes, (list, tuple)) else [next_nodes]
        self.next_nodes += next_nodes
        self.next_node_names += [ii.name for ii in next_nodes]


class Shape(GraphNode):
    """
    >>> from keras_cv_attention_models.pytorch_backend import layers, functional
    >>> aa = layers.Input([4, 4, 32])
    >>> bb = functional.shape(aa)
    >>> print(bb)
    >>> # [-1, 4, 4, 32]
    >>> aa.set_shape([None, 3, 4, 5])
    >>> print(bb)
    >>> # [-1, 3, 4, 5]
    """

    num_instances = 0  # Count instances

    @classmethod
    def __count__(cls):
        cls.num_instances += 1

    def __init__(self, input_node, index_expr=slice(None), name=None):
        self.name = "shape_{}".format(self.num_instances) if name is None else name
        self.input_node, self.index_expr = input_node, index_expr
        self.is_slice = not isinstance(self.index_expr, int)
        self.shape = (self.__len__(),) if self.is_slice else ()
        self.build()
        self.__count__()

    def build(self):
        if self.is_slice:
            self.callable = lambda inputs: [-1 if ii is None else ii for ii in inputs.shape[self.index_expr]]
        else:
            self.callable = lambda inputs: inputs.shape[self.index_expr]
        self.layer = self
        self.pre_nodes = [self.input_node]
        self.pre_node_names = [self.input_node.name]
        self.input_node.set_next_nodes(self)
        self.weights = []
        self.next_nodes, self.next_node_names = [], []

    def set_next_nodes(self, next_nodes):
        next_nodes = next_nodes if isinstance(next_nodes, (list, tuple)) else [next_nodes]
        self.next_nodes += next_nodes
        self.next_node_names += [ii.name for ii in next_nodes]

    def get_weights(self):
        return []

    def __getitem__(self, index_expr):
        return Shape(self.input_node, index_expr)

    def __index__(self):
        return self.input_node.shape[self.index_expr]

    def __value__(self):
        return [-1 if ii is None else ii for ii in self.input_node.shape[self.index_expr]] if self.is_slice else (self.input_node.shape[self.index_expr],)

    def __len__(self):
        return len(self.__value__())  # if self.is_slice else 1

    def __int__(self):
        return int(self.__value__()[0])

    def __iter__(self):
        return (ii for ii in self.__value__())  # if self.is_slice else (self.__value__(),)

    def __repr__(self):
        return str(self.__value__())


class Input(GraphNode):
    def __init__(self, shape, name=None, dtype=None):
        shape = [None, *shape]
        name = "input_{}".format(self.num_instances) if name is None else name
        super().__init__(shape, name=name)
        self.dtype = torch.get_default_dtype() if dtype is None else (getattr(torch, dtype) if isinstance(dtype, str) else dtype)


class Layer(nn.Module):
    num_instances = 0  # Count instances

    @classmethod
    def __count__(cls):
        cls.num_instances += 1

    def __init__(self, name=None, **kwargs):
        super().__init__()
        self.name, self.kwargs = self.verify_name(name), kwargs
        self.built = False
        # self.is_graph_node_input = False
        self.use_layer_as_module = False  # True if called add_weights, or typical keras layers with overwriting call function
        if not hasattr(self, "module"):
            self.module = lambda xx: xx
        self.__count__()
        self.outputs = self.nodes = None

    def build(self, input_shape: torch.Size):
        # pass
        # self.input_shape = input_shape
        # self.dtype = self.module.weight.dtype if hasattr(self.module, "weight") else "float32"
        # if self.is_graph_node_input:  # When exporting onnx/pth, model will be built again, and may throws error in compute_output_shape with 0 input_shape
        # self.__output_shape__ = self.compute_output_shape(input_shape)
        # if hasattr(self, "call"):  # General keras layers with call function
        #     self.forward = self.call
        #     self.call = self.call
        self.built = True

    def call(self, inputs, **kwargs):
        return self.module(inputs, **kwargs)

    def forward(self, inputs, **kwargs):
        # self.is_graph_node_input = isinstance(inputs, GraphNode) or (isinstance(inputs, (list, tuple)) and any([isinstance(ii, GraphNode) for ii in inputs]))
        if not self.built:
            self.build([() if isinstance(ii, (int, float)) else ii.shape for ii in inputs] if isinstance(inputs, (list, tuple)) else inputs.shape)
            self.built = True

        if isinstance(inputs, GraphNode) or (isinstance(inputs, (list, tuple)) and any([isinstance(ii, GraphNode) for ii in inputs])):
            self.input_shape = [() if isinstance(ii, (int, float)) else ii.shape for ii in inputs] if isinstance(inputs, (list, tuple)) else inputs.shape
            self.__output_shape__ = self.compute_output_shape(self.input_shape)

            # output_shape = self.compute_output_shape(input_shape)
            # if isinstance(self.__output_shape__[0], (list, tuple))
            cur_node = GraphNode(self.__output_shape__, name=self.name if self.nodes is None else (self.name + "_{}".format(len(self.nodes))))
            if self.use_layer_as_module:  # General keras layers with call function, mostly own weights
                cur_node.callable = self
            else:
                cur_node.callable = self.module if len(kwargs) == 0 else partial(self.module, **kwargs)
            cur_node.layer = self
            cur_node.set_pre_nodes(inputs)

            inputs = inputs if isinstance(inputs, (list, tuple)) else [inputs]
            for ii in inputs:
                if isinstance(ii, GraphNode):
                    ii.set_next_nodes(cur_node)
            self.inputs = inputs
            self.input = inputs[0] if len(inputs) == 1 else inputs

            if self.nodes is None:
                self.outputs = self.nodes = [cur_node]
                self.output = self.node = cur_node
            else:
                self.nodes.append(cur_node)
                self.output = self.node = self.outputs = self.nodes
            return cur_node
        else:
            return self.call(inputs, **kwargs)

    @property
    def compute_dtype(self):
        try:
            return next(self.parameters()).dtype
        except StopIteration:
            return torch.get_default_dtype()

    @property
    def weights(self):
        return [Weight(name=self.name + "/" + kk.split(".")[-1], value=vv) for kk, vv in self.state_dict().items() if not kk.endswith(".num_batches_tracked")]

    @property
    def trainable_weights(self):
        return self.weights

    @property
    def non_trainable_weights(self):
        return []

    def add_weight(self, name=None, shape=None, dtype="float32", initializer="zeros", regularizer=None, trainable=True):
        self.use_layer_as_module = True
        initializer = getattr(initializers, initializer)() if isinstance(initializer, str) else initializer
        param = nn.Parameter(initializer(shape, dtype=dtype), requires_grad=trainable)
        return param

    def register_buffer(self, name, tensor, persistent=True):
        self.use_layer_as_module = True
        super().register_buffer(name=name, tensor=tensor, persistent=persistent)

    def get_weights(self):
        return [ii.value().detach().cpu().numpy() for ii in self.weights]

    def set_weights(self, weights):
        return self.load_state_dict({kk: torch.from_numpy(vv) for kk, vv in zip(self.state_dict().keys(), weights)})

    def compute_output_shape(self, input_shape):
        return input_shape

    @property
    def output_shape(self):
        return getattr(self, "__output_shape__", None)

    def verify_name(self, name):
        return "{}_{}".format(self.__class__.__name__.lower(), self.num_instances) if name == None else name

    def get_config(self):
        config = {"name": self.name}
        config.update(self.kwargs)
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def extra_repr(self):
        config = self.get_config()
        # config.pop("name")
        return ", ".join(["{}={}".format(kk, vv) for kk, vv in config.items()])


class Lambda(Layer):
    def __init__(self, func, output_shape=None, **kwargs):
        self.module = func
        self.__output_shape__ = output_shape
        super().__init__(**kwargs)

    def build(self, input_shape: torch.Size):
        super().build(input_shape)
        # self.forward = self.module

    def get_config(self):
        config = super().get_config()
        config.update({"func": self.module, "output_shape": self.__output_shape__})
        return config

    def compute_output_shape(self, input_shape):
        if self.__output_shape__ is not None:
            return self.__output_shape__

        # print(self.module, input_shape)
        input_shapes = [input_shape] if isinstance(input_shape[0], int) or input_shape[0] is None else input_shape
        # print(input_shapes)
        inputs = [torch.ones([0 if ii is None or ii == -1 else ii for ii in input_shape]) for input_shape in input_shapes]  # Regards 0 as dynamic shape
        output_shape = list(self.module(inputs[0]).shape) if len(inputs) == 1 else list(self.module(inputs).shape)
        return [None if ii == 0 else ii for ii in output_shape]  # Regards 0 as dynamic shape

    def extra_repr(self):
        from types import LambdaType

        if isinstance(self.module, partial) and isinstance(self.module.func, LambdaType):
            func_str = "{}, {}".format(inspect.getsource(self.module.func), self.module.keywords)
        if isinstance(self.module, partial):
            func_str = "{}, {}".format(self.module.func.__name__, self.module.keywords)
        else:
            func_str = inspect.getsource(self.module)
        return "name: " + self.name + ", " + func_str.strip()


class Activation(Layer):
    def __init__(self, activation=None, **kwargs):
        self.activation = activation
        super().__init__(**kwargs)

    def build(self, input_shape):
        if self.activation is None or self.activation == "linear":
            self.module = torch.nn.Identity()
        elif isinstance(self.activation, str) and self.activation == "softmax":
            self.module = partial(torch.softmax, dim=1)
        elif isinstance(self.activation, str) and self.activation == "swish":
            self.module = torch.nn.SiLU()
        elif isinstance(self.activation, str):
            self.module = getattr(torch, self.activation, getattr(torch.functional.F, self.activation))
        else:
            self.module = self.activation
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"activation": self.activation})
        return config


class Identity(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = torch.nn.Identity()
        super().build(input_shape)


""" Merge Layers """


class _Merge(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.check_input(list(input_shape))
        self.module = self.merge_function
        super().build(input_shape)

    def preprocess_input_shape(self, input_shape):
        valid_input_shape = [ii for ii in input_shape if len(ii) != 0]  # exclude scalar values, also cut batch dimension
        if len(valid_input_shape) == 0:  # Should not happen...
            return (), ()

        max_dim = max([len(ii) for ii in valid_input_shape]) - 1
        valid_input_shape = [ii if len(ii) == 1 else ii[1:] for ii in valid_input_shape]  # Regard None as 0
        valid_input_shape = [[1] * (max_dim - len(ii)) + list(ii) for ii in valid_input_shape]  # expands 1 on the head
        valid_input_shape = np.array([[np.inf if jj is None else jj for jj in ii] for ii in valid_input_shape])  # Regard None as inf
        max_shape = valid_input_shape.max(0)
        max_shape = [None if np.isinf(ii) else int(ii) for ii in max_shape]  # Regard inf as None
        return valid_input_shape, max_shape

    def check_input(self, input_shape):
        valid_input_shape, max_shape = self.preprocess_input_shape(input_shape)
        if len(valid_input_shape) == 0:
            return
        check_single_func = lambda ii: all([ss is None or tt is None or ss == 1 or ss == tt for ss, tt in zip(ii, max_shape)])
        assert all([check_single_func(ii) for ii in valid_input_shape]), "input_shapes not all equal: {}".format(input_shape)

    def compute_output_shape(self, input_shape):
        valid_input_shape, max_shape = self.preprocess_input_shape(input_shape)
        return () if len(valid_input_shape) == 0 else [None, *max_shape]


class Add(_Merge):
    def merge_function(self, inputs):
        output = torch.add(inputs[0], inputs[1])
        for ii in inputs[2:]:
            output = torch.add(output, ii)
        return output


class Divide(_Merge):
    def merge_function(self, inputs):
        output = torch.divide(inputs[0], inputs[1])
        for ii in inputs[2:]:
            output = torch.divide(output, ii)
        return output


class Equal(_Merge):
    def merge_function(self, inputs):
        output = torch.eq(inputs[0], inputs[1])
        return output


class Subtract(_Merge):
    def merge_function(self, inputs):
        output = torch.subtract(inputs[0], inputs[1])
        for ii in inputs[2:]:
            output = torch.subtract(output, ii)
        return output


class Matmul(_Merge):
    def merge_function(self, inputs):
        output = torch.matmul(inputs[0], inputs[1])
        for ii in inputs[2:]:
            output = torch.matmul(output, ii)
        return output

    def check_input(self, input_shape):
        base, dims = input_shape[0], len(input_shape[0])
        result = not any([any([base[dim] != ii[dim] for dim in range(1, dims) if dim != self.axis]) for ii in input_shape[1:]])
        assert result, "input_shapes excpets concat axis {} not all equal: {}".format(self.axis, input_shape)

    def compute_output_shape(self, input_shape):
        dims = len(input_shape[0])
        return [sum([ii[dim] for ii in input_shape]) if dim == self.axis else input_shape[0][dim] for dim in range(dims)]


class Multiply(_Merge):
    def merge_function(self, inputs):
        output = torch.multiply(inputs[0], inputs[1])
        for ii in inputs[2:]:
            output = torch.multiply(output, ii)
        return output


class Concatenate(_Merge):
    def __init__(self, axis=1, **kwargs):
        self.axis = axis
        super().__init__(**kwargs)

    def merge_function(self, inputs):
        return torch.concat(inputs, dim=self.axis)

    def check_input(self, input_shape):
        base, dims = input_shape[0], len(input_shape[0])
        axis = (dims + self.axis) if self.axis < 0 else self.axis
        result = not any([any([base[dim] != ii[dim] for dim in range(1, dims) if dim != axis]) for ii in input_shape[1:]])
        assert result, "input_shapes excpets concat axis {} not all equal: {}".format(self.axis, input_shape)

    def compute_output_shape(self, input_shape):
        output_shape = list(input_shape[0]).copy()
        combine_dims = [ii[self.axis] for ii in input_shape]
        output_shape[self.axis] = None if any([ii is None for ii in combine_dims]) else sum(combine_dims)
        return output_shape
        # dims = len(input_shape[0])
        # axis = (dims + self.axis) if self.axis < 0 else self.axis
        # return [sum([ii[dim] for ii in input_shape]) if dim == axis else input_shape[0][dim] for dim in range(dims)]

    def get_config(self):
        config = super().get_config()
        config.update({"axis": self.axis})
        return config


""" Layers with weights """


class BatchNormalization(Layer):
    def __init__(self, axis=1, momentum=0.9, epsilon=1e-5, center=True, gamma_initializer="ones", **kwargs):
        self.axis, self.momentum, self.epsilon, self.center, self.gamma_initializer = axis, momentum, epsilon, center, gamma_initializer
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.axis = len(input_shape) + self.axis if self.axis < 0 else self.axis
        module = nn.BatchNorm2d(num_features=input_shape[self.axis], eps=self.epsilon, momentum=1 - self.momentum, affine=self.center)
        gamma_initializer = getattr(initializers, self.gamma_initializer)() if isinstance(self.gamma_initializer, str) else self.gamma_initializer
        module.weight.data = gamma_initializer(list(module.weight.shape))  # not using gamma_initializer(module.weight) for compiling with TF

        if self.axis == 1:
            self.module = module
        else:
            ndims = len(input_shape)
            perm = get_perm(total_axis=ndims, from_axis=self.axis, to_axis=1)  # like axis=-1 -> [0, 3, 1, 2]
            revert_perm = get_perm(total_axis=ndims, from_axis=1, to_axis=self.axis)  # like axis=-1 -> [0, 2, 3, 1]
            self.module = nn.Sequential(Permute(perm[1:]), module, Permute(revert_perm[1:]))
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"axis": self.axis, "momentum": self.momentum, "epsilon": self.epsilon, "center": self.center})
        return config


class _SamePadding(nn.Module):
    """Perform same padding like TF"""

    def __init__(self, kernel_size, strides, dilation_rate=1, ndims=2, value=0):
        super().__init__()
        self.kernel_size = kernel_size if isinstance(kernel_size, (list, tuple)) else [kernel_size] * ndims
        self.strides = strides if isinstance(strides, (list, tuple)) else [strides] * ndims
        self.dilation_rate = dilation_rate if isinstance(dilation_rate, (list, tuple)) else [dilation_rate] * ndims
        self.ndims = len(self.kernel_size)
        self.value = value  # For MaxPool2D, needs to be -inf
        self.fixed_padding = None

    def build_pad(self, input_shape):
        pad = [tf_same_pad(ii, kk, ss, dd) for ii, kk, ss, dd in zip(input_shape[2:], self.kernel_size, self.strides, self.dilation_rate)]
        padding = []
        for pp in pad[::-1]:
            padding += [pp // 2, pp - pp // 2]
        padding += [0, 0, 0, 0]
        return padding

    def forward(self, inputs):
        padding = self.build_pad(inputs.shape) if self.fixed_padding is None else self.fixed_padding
        return torch.functional.F.pad(inputs, pad=padding, value=self.value)


class _ZeroPadding(Layer):
    def __init__(self, padding, mode="constant", constant_values=None, **kwargs):
        self.padding, self.mode, self.constant_values = padding, mode, constant_values
        super().__init__(**kwargs)

    def build(self, input_shape):
        # print(f"{input_shape = }, {self.padding = }")
        if self.mode == "replicate":
            assert len(input_shape) > len(self.padding), "replicate padding shoule not including batch dimension, padding: {}".format(self.padding)
        else:
            assert len(input_shape) == len(self.padding), "padding shoule be same length with input, including batch, padding: {}".format(self.padding)
        padding = []
        for pad in self.padding[::-1]:
            assert len(pad) == 2, "each element in padding should be exactly 2 values, padding: {}".format(self.padding)
            padding += pad
        # print(f"{padding = }")

        self.module = partial(torch.functional.F.pad, pad=padding, mode=self.mode, value=self.constant_values)
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        if self.mode == "replicate":
            pad_len = len(self.padding)
            return input_shape[:-pad_len] + [None if ii is None else ii + pp[0] + pp[1] for ii, pp in zip(input_shape[-pad_len:], self.padding)]
        else:
            return [None if ii is None else ii + pp[0] + pp[1] for ii, pp in zip(input_shape, self.padding)]

    def get_config(self):
        config = super().get_config()
        config.update({"padding": self.padding, "mode": self.mode, "constant_values": self.constant_values})
        return config


class _BaseConvPool(Layer):
    def __init__(self, kernel_size=1, strides=1, padding="valid", dilation_rate=1, padding_value=0, **kwargs):
        self.kernel_size, self.dilation_rate, self.strides, self.padding, self.padding_value = kernel_size, dilation_rate, strides, padding, padding_value
        super().__init__(**kwargs)

    def build_module(self, input_shape):
        raise NotImplementedError()

    def build(self, input_shape):
        num_dims = len(input_shape) - 2  # Conv2D -> 2, Conv1D -> 1
        self.kernel_size = to_tuple(self.kernel_size, num_dims=num_dims)
        self.dilation_rate = to_tuple(self.dilation_rate, num_dims=num_dims)
        self.strides = to_tuple(self.strides, num_dims=num_dims)

        # elif isinstance(self.padding, str):
        if isinstance(self.padding, str):
            self._pad = [0] * num_dims  # Alo set to 0 for "same" if input_shape[-1] is None, will apply TF like same padding later.
            padding = self.padding.lower()
        else:  # int or list or tuple with specific value
            self._pad = self.padding if isinstance(self.padding, (list, tuple)) else [self.padding] * num_dims
            padding = self.padding

        if None not in input_shape[2:] and padding == "same":
            pad = [tf_same_pad(*args) for args in zip(input_shape[2:], self.kernel_size, self.strides, self.dilation_rate)]
            half_pad = [pp // 2 for pp in pad]
            if all([ii == 2 * jj for ii, jj in zip(pad, half_pad)]):
                self._pad = half_pad  # Using module pad directly, avoid applying additional `_SamePadding` layer
                padding = "valid"

        module = self.build_module(input_shape)
        if max(self.kernel_size) > 0 and padding == "same":
            # TF like same padding
            same_padding = _SamePadding(self.kernel_size, strides=self.strides, dilation_rate=self.dilation_rate, ndims=num_dims, value=self.padding_value)
            if None not in input_shape[1:]:
                same_padding.fixed_padding = same_padding.build_pad(input_shape)  # Set fixed padding
            self.module = nn.Sequential(same_padding, module)
        else:
            self.module = module
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        output_size = compute_conv_output_size(input_shape, self.kernel_size, self.strides, self.padding, dilation_rate=self.dilation_rate)
        return [None, input_shape[1], *output_size]

    def get_config(self):
        config = super().get_config()
        config.update({"kernel_size": self.kernel_size, "strides": self.strides, "padding": self.padding, "dilation_rate": self.dilation_rate})
        # config.update({"padding_value": self.padding_value})
        return config


class Conv(_BaseConvPool):
    def __init__(
        self,
        filters,
        kernel_size=1,
        strides=1,
        padding="valid",
        dilation_rate=1,
        use_bias=True,
        groups=1,
        kernel_initializer=None,
        bias_initializer="zeros",
        **kwargs,
    ):
        super().__init__(kernel_size=kernel_size, strides=strides, padding=padding, dilation_rate=dilation_rate, **kwargs)
        self.filters, self.use_bias, self.groups = filters, use_bias, groups
        self.kernel_initializer, self.bias_initializer = kernel_initializer, bias_initializer
        self.module_class = None  # Auto set by len(input_shape)

    def build_module(self, input_shape):
        self.filters = int(self.filters if self.filters > 0 else input_shape[1])  # In case DepthwiseConv2D
        if self.module_class is None:
            self.module_class = nn.Conv1d if len(input_shape) == 3 else (nn.Conv2d if len(input_shape) == 4 else nn.Conv3d)

        module = self.module_class(
            in_channels=input_shape[1],
            out_channels=self.filters,
            kernel_size=self.kernel_size,
            stride=self.strides,
            padding=self._pad,
            dilation=self.dilation_rate,
            groups=self.groups,
            bias=self.use_bias,
        )
        if self.kernel_initializer is not None:
            kernel_initializer = getattr(initializers, self.kernel_initializer)() if isinstance(self.kernel_initializer, str) else self.kernel_initializer
            module.weight.data = kernel_initializer(list(module.weight.shape))  # not using kernel_initializer(module.weight) for compiling with TF
        if self.use_bias and self.bias_initializer is not None:
            bias_initializer = getattr(initializers, self.bias_initializer)() if isinstance(self.bias_initializer, str) else self.bias_initializer
            module.bias.data = bias_initializer(list(module.bias.shape))
        return module

    def compute_output_shape(self, input_shape):
        output_size = compute_conv_output_size(input_shape, self.kernel_size, self.strides, self.padding, dilation_rate=self.dilation_rate)
        return [None, self.filters, *output_size]

    def get_config(self):
        config = super().get_config()
        config.update({"filters": self.filters, "use_bias": self.use_bias, "groups": self.groups})
        return config


class Conv1D(Conv):
    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (2, 1, 0))
        return self.set_weights(weights)


class Conv2D(Conv):
    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (3, 2, 0, 1))
        return self.set_weights(weights)


class Conv3D(Conv):
    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 4, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (4, 3, 0, 1, 2))
        return self.set_weights(weights)


class ConvTranspose(_BaseConvPool):
    def __init__(
        self,
        filters,
        kernel_size=1,
        strides=1,
        padding=0,
        output_padding=None,
        dilation_rate=1,
        use_bias=True,
        kernel_initializer=None,
        bias_initializer="zeros",
        **kwargs,
    ):
        super().__init__(kernel_size=kernel_size, strides=strides, padding=padding, dilation_rate=dilation_rate, **kwargs)
        self.filters, self.use_bias, self.output_padding = int(filters), use_bias, output_padding
        self.kernel_initializer, self.bias_initializer = kernel_initializer, bias_initializer
        self.output_padding = to_tuple(output_padding, num_dims=2)
        self.module_class = None

    def build_module(self, input_shape):
        if self.module_class is None:
            self.module_class = nn.ConvTranspose1d if len(input_shape) == 3 else (nn.ConvTranspose2d if len(input_shape) == 4 else nn.ConvTranspose3d)

        module = self.module_class(
            in_channels=input_shape[1],
            out_channels=self.filters,
            kernel_size=self.kernel_size,
            stride=self.strides,
            padding=self._pad,
            output_padding=0 if self.output_padding[0] is None else self.output_padding,
            dilation=self.dilation_rate,
            bias=self.use_bias,
        )
        if self.kernel_initializer is not None:
            kernel_initializer = getattr(initializers, self.kernel_initializer)() if isinstance(self.kernel_initializer, str) else self.kernel_initializer
            module.weight.data = kernel_initializer(list(module.weight.shape))
        if self.use_bias and self.bias_initializer is not None:
            bias_initializer = getattr(initializers, self.bias_initializer)() if isinstance(self.bias_initializer, str) else self.bias_initializer
            module.bias.data = bias_initializer(list(module.bias.shape))

        return module

    def deconv_output_length(self, size, kernel_size, strides=1, pad=0, output_padding=None, dilation=1):
        # keras.utils.conv_utils.deconv_output_length
        if size is None:
            return None

        # Get the dilated kernel size
        kernel_size = kernel_size + (kernel_size - 1) * (dilation - 1)

        # Infer length if output padding is None, else compute the exact length
        if output_padding is None:
            length = size * strides
            if pad == 0:  # "valid"
                length += max(kernel_size - strides, 0)
        else:
            length = (size - 1) * strides + kernel_size
            if padding > 0:
                length += kernel_size // 2 * 2 + output_padding
        return length

    def compute_output_shape(self, input_shape):
        input_size = input_shape[2:]
        out = [self.deconv_output_length(*args) for args in zip(input_size, self.kernel_size, self.strides, self._pad, self.output_padding, self.dilation_rate)]
        return [None, self.filters, *out]

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (3, 2, 0, 1))
        return self.set_weights(weights)

    def get_config(self):
        config = super().get_config()
        config.update({"filters": self.filters, "use_bias": self.use_bias, "output_padding": self.output_padding})
        return config


class Conv1DTranspose(ConvTranspose):
    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (2, 1, 0))
        return self.set_weights(weights)


class Conv2DTranspose(ConvTranspose):
    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (3, 2, 0, 1))
        return self.set_weights(weights)


class Conv3DTranspose(ConvTranspose):
    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 4, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (4, 3, 0, 1, 2))
        return self.set_weights(weights)


class DepthwiseConv2D(Conv):
    def __init__(self, kernel_size=1, strides=1, padding="valid", dilation_rate=(1, 1), use_bias=True, depthwise_initializer="glorot_uniform", **kwargs):
        filters, groups = -1, -1
        super().__init__(filters, kernel_size, strides, padding, dilation_rate, use_bias, groups, depthwise_initializer, **kwargs)

    def build(self, input_shape):
        self.groups = input_shape[1]
        super().build(input_shape)

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 0, 1))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (2, 3, 0, 1))
        return self.set_weights(weights)


class Dense(Layer):
    def __init__(self, units, activation=None, use_bias=True, axis=-1, kernel_initializer=None, bias_initializer="zeros", **kwargs):
        self.units, self.activation, self.use_bias, self.axis = int(units), activation, use_bias, axis
        self.kernel_initializer, self.bias_initializer = kernel_initializer, bias_initializer
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.axis = len(input_shape) + self.axis if self.axis < 0 else self.axis
        module = nn.Linear(in_features=input_shape[self.axis], out_features=self.units, bias=self.use_bias)

        if self.kernel_initializer is not None:
            kernel_initializer = getattr(initializers, self.kernel_initializer)() if isinstance(self.kernel_initializer, str) else self.kernel_initializer
            module.weight.data = kernel_initializer(list(module.weight.shape))  # not using kernel_initializer(module.weight) for compiling with TF
        if self.use_bias and self.bias_initializer is not None:
            bias_initializer = getattr(initializers, self.bias_initializer)() if isinstance(self.bias_initializer, str) else self.bias_initializer
            module.bias.data = bias_initializer(list(module.bias.shape))

        if self.axis == len(input_shape) - 1:
            self.module = module if self.activation is None else nn.Sequential(module, Activation(self.activation))
        else:
            ndims = len(input_shape)
            perm = get_perm(total_axis=ndims, from_axis=self.axis, to_axis=-1)  # like axis=1 -> [0, 2, 3, 1]
            revert_perm = get_perm(total_axis=ndims, from_axis=-1, to_axis=self.axis)  # like axis=1 -> [0, 3, 1, 2]
            if self.activation is None:
                self.module = nn.Sequential(Permute(perm[1:]), module, Permute(revert_perm[1:]))
            else:
                self.module = nn.Sequential(Permute(perm[1:]), module, Permute(revert_perm[1:]), Activation(self.activation))
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        input_shape = list(input_shape)
        return input_shape[: self.axis] + [self.units] + input_shape[self.axis + 1 :]

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0])
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0])
        return self.set_weights(weights)

    def get_config(self):
        config = super().get_config()
        config.update({"units": self.units, "activation": self.activation, "use_bias": self.use_bias, "axis": self.axis})
        return config


class Embedding(Layer):
    def __init__(self, input_dim, output_dim, embeddings_initializer="random_uniform", mask_zero=False, input_length=None, **kwargs):
        self.input_dim, self.output_dim, self.mask_zero, self.input_length = input_dim, output_dim, mask_zero, input_length
        self.embeddings_initializer = embeddings_initializer
        super().__init__(**kwargs)

    def build(self, input_shape):
        padding_idx = 0 if self.mask_zero else None
        self.module = torch.nn.Embedding(num_embeddings=self.input_dim, embedding_dim=self.output_dim, padding_idx=padding_idx, max_norm=None)
        initializer = getattr(initializers, self.embeddings_initializer)() if isinstance(self.embeddings_initializer, str) else self.embeddings_initializer
        self.module.weight.data = initializer(list(self.module.weight.shape))
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        input_shape = list(input_shape)
        return input_shape + [self.output_dim]

    def get_config(self):
        config = super().get_config()
        config.update({"input_dim": self.input_dim, "output_dim": self.output_dim, "mask_zero": self.mask_zero, "input_length": self.input_length})
        return config


class GroupNormalization(Layer):
    def __init__(self, groups=32, axis=1, epsilon=0.001, center=True, gamma_initializer="ones", **kwargs):
        self.groups, self.axis, self.epsilon, self.center, self.gamma_initializer = groups, axis, epsilon, center, gamma_initializer
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.axis = len(input_shape) + self.axis if self.axis < 0 else self.axis
        module = nn.GroupNorm(num_groups=self.groups, num_channels=input_shape[self.axis], eps=self.epsilon, affine=self.center)
        gamma_initializer = getattr(initializers, self.gamma_initializer)() if isinstance(self.gamma_initializer, str) else self.gamma_initializer
        module.weight.data = gamma_initializer(list(module.weight.shape))  # not using gamma_initializer(module.weight) for compiling with TF
        # Default nn.GroupNorm is apllied on first non-batch dimension
        if self.axis == 1:
            self.module = module
        else:
            ndims = len(input_shape)
            perm = get_perm(total_axis=ndims, from_axis=self.axis, to_axis=1)  # like axis=-1 -> [0, 3, 1, 2]
            revert_perm = get_perm(total_axis=ndims, from_axis=1, to_axis=self.axis)  # like axis=-1 -> [0, 2, 3, 1]
            self.module = nn.Sequential(Permute(perm[1:]), module, Permute(revert_perm[1:]))
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"groups": self.groups, "axis": self.axis, "epsilon": self.epsilon, "center": self.center})
        return config


class _LayerNormGeneral(nn.Module):
    """LayerNorm supports `bias=False`, also applying on `axis=1` directly without permute.
    From LayerNormGeneral https://github.com/sail-sg/metaformer/blob/main/metaformer_baselines.py#L311
    """

    def __init__(self, normalized_shape, normalized_dim=(-1,), scale=True, bias=True, eps=1e-5):
        super().__init__()
        self.normalized_dim, self.use_scale, self.use_bias, self.eps = normalized_dim, scale, bias, eps
        self.weight = nn.Parameter(torch.ones(normalized_shape)) if scale else None
        self.bias = nn.Parameter(torch.zeros(normalized_shape)) if bias else None

    def forward(self, inputs):
        center = inputs - inputs.mean(self.normalized_dim, keepdim=True)
        scale = center.pow(2).mean(self.normalized_dim, keepdim=True)
        inputs = center / torch.sqrt(scale + self.eps)
        if self.use_scale:
            inputs = inputs * self.weight
        if self.use_bias:
            inputs = inputs + self.bias
        return inputs


class LayerNormalization(Layer):
    def __init__(self, axis=1, epsilon=1e-5, center=True, gamma_initializer="ones", **kwargs):
        self.axis, self.epsilon, self.center, self.gamma_initializer = axis, epsilon, center, gamma_initializer
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.axis = len(input_shape) + self.axis if self.axis < 0 else self.axis
        if self.axis == len(input_shape) - 1 and self.center:
            self.module = nn.LayerNorm(normalized_shape=input_shape[self.axis], eps=self.epsilon)
            gamma_initializer = getattr(initializers, self.gamma_initializer)() if isinstance(self.gamma_initializer, str) else self.gamma_initializer
            self.module.weight.data = gamma_initializer(list(self.module.weight.shape))
            # Default nn.LayerNorm is apllied on last dimension
            # if self.axis != len(input_shape) - 1:
            #     ndims = len(input_shape)
            #     perm = get_perm(total_axis=ndims, from_axis=self.axis, to_axis=-1)  # like axis=1 -> [0, 2, 3, 1]
            #     revert_perm = get_perm(total_axis=ndims, from_axis=-1, to_axis=self.axis)  # like axis=1 -> [0, 3, 1, 2]
            #     self.module = nn.Sequential(Permute(perm[1:]), self.module, Permute(revert_perm[1:]))
        else:
            normalized_shape = input_shape[self.axis]
            if self.axis != len(input_shape) - 1:
                normalized_shape = [normalized_shape] + [1] * (len(input_shape) - self.axis - 1)
            self.module = _LayerNormGeneral(normalized_shape=normalized_shape, normalized_dim=self.axis, eps=self.epsilon, bias=self.center)
            gamma_initializer = getattr(initializers, self.gamma_initializer)() if isinstance(self.gamma_initializer, str) else self.gamma_initializer
            self.module.weight.data = gamma_initializer(list(self.module.weight.shape))  # not using gamma_initializer(self.module.weight) for compiling with TF
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"axis": self.axis, "epsilon": self.epsilon, "center": self.center})
        return config

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights = [np.squeeze(ww) for ww in weights]
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights = [np.reshape(ss, tt.shape) for ss, tt in zip(weights, self.weights)]
        return self.set_weights(weights)


class PReLU(Layer):
    def __init__(self, alpha_initializer="zeros", alpha_regularizer=None, alpha_constraint=None, shared_axes=None, **kwargs):
        self.alpha_initializer, self.shared_axes = alpha_initializer, shared_axes
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = nn.PReLU(num_parameters=input_shape[1], init=0.25)
        alpha_initializer = getattr(initializers, self.alpha_initializer)() if isinstance(self.alpha_initializer, str) else self.alpha_initializer
        self.module.weight.data = alpha_initializer(list(self.module.weight.shape))  # not using alpha_initializer(self.module.weight) for compiling with TF
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"shared_axes": self.shared_axes})
        return config

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.expand_dims(np.expand_dims(weights[0], 0), 0)
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.squeeze(weights[0])
        return self.set_weights(weights)


class SeparableConv2D(Conv):
    def __init__(
        self,
        filters,
        kernel_size=1,
        strides=1,
        padding="valid",
        dilation_rate=1,
        use_bias=True,
        pointwise_initializer=None,
        bias_initializer="zeros",
        **kwargs,
    ):
        groups = 1
        super().__init__(filters, kernel_size, strides, padding, dilation_rate, use_bias, groups, pointwise_initializer, bias_initializer, **kwargs)

    def build_module(self, input_shape):
        depthwise = nn.Conv2d(
            in_channels=input_shape[1],
            out_channels=input_shape[1],
            kernel_size=self.kernel_size,
            stride=self.strides,
            padding=self._pad,
            dilation=self.dilation_rate,
            groups=input_shape[1],
            bias=False,
        )
        pointwise = nn.Conv2d(
            in_channels=input_shape[1],
            out_channels=self.filters,
            kernel_size=1,
            stride=1,
            padding=0,
            dilation=1,
            groups=1,
            bias=self.use_bias,
        )

        if self.kernel_initializer is not None:
            kernel_initializer = getattr(initializers, self.kernel_initializer)() if isinstance(self.kernel_initializer, str) else self.kernel_initializer
            depthwise.weight.data = kernel_initializer(list(depthwise.weight.shape))
            pointwise.weight.data = kernel_initializer(list(pointwise.weight.shape))
        if self.use_bias and self.bias_initializer is not None:
            bias_initializer = getattr(initializers, self.bias_initializer)() if isinstance(self.bias_initializer, str) else self.bias_initializer
            pointwise.bias.data = bias_initializer(list(pointwise.bias.shape))

        return nn.Sequential(depthwise, pointwise)

    def get_weights_channels_last(self):
        # channel_first -> channel_last
        weights = self.get_weights()
        weights[0] = np.transpose(weights[0], (2, 3, 0, 1))
        weights[1] = np.transpose(weights[1], (2, 3, 1, 0))
        return weights

    def set_weights_channels_last(self, weights):
        # channel_last -> channel_first
        weights[0] = np.transpose(weights[0], (2, 3, 0, 1))
        weights[1] = np.transpose(weights[1], (3, 2, 0, 1))
        return self.set_weights(weights)


""" Layers with no weights """


class _GatherND(Layer):
    """Works like `tf.gather_nd`
    Resize supporting dynamic size.
    >>> import torch
    >>> from keras_cv_attention_models.pytorch_backend import layers
    >>> print(layers._GatherND(batch_dims=0)([layers.Input([3, 4]), torch.arange(2)[:, None]]).shape)
    >>> # [2, 3, 4]
    >>> aa = torch.arange(12).reshape([3, 4, 1])
    >>> print(layers._GatherND(batch_dims=1)([aa, torch.arange(3)[:, None]]).shape)
    >>> # torch.Size([3, 1])
    """

    def __init__(self, batch_dims=0, **kwargs):
        self.batch_dims = batch_dims
        super().__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 2, "Input should be in format [inputs, indices]"
        input_shape, indices_shape = input_shape[0], input_shape[1]

        if len(indices_shape) > 1:
            assert (
                len(input_shape) >= self.batch_dims + indices_shape[-1]
            ), f"Provided input_shape={input_shape} length should be larger than indices_shape[-1]={indices_shape[-1]} + batch_dims={self.batch_dims}"
        if self.batch_dims > 0:
            assert self.batch_dims < len(indices_shape), f"Argument `batch_dims` = {self.batch_dims} must be less than rank(`indices`) = {len(indices_shape)}"

        if len(indices_shape) > 1 and self.batch_dims > 0:
            self.module = lambda inputs: inputs[0][[torch.arange(inputs[0].shape[ii]) for ii in range(self.batch_dims)] + inputs[1].T.tolist()]
            self.gathered_indices = list(range(self.batch_dims, self.batch_dims + indices_shape[-1]))
        elif len(indices_shape) > 1:
            self.module = lambda inputs: inputs[0][inputs[1].T.tolist()]
            self.gathered_indices = list(range(indices_shape[-1]))
        else:  # Gather on batch dimension
            self.module = lambda inputs: inputs[0][tuple(inputs[1].tolist())]  # .contiguous()
            self.gathered_indices = [0]
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        batch_dimension = input_shape[1][0] if 0 in self.gathered_indices else None
        return [batch_dimension] + [ii for id, ii in enumerate(input_shape[0][1:], start=1) if id not in self.gathered_indices]

    def get_config(self):
        config = super().get_config()
        config.update({"batch_dims": self.batch_dims})
        return config


class _DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    from drop_path https://github.com/rwightman/pytorch-image-models/blob/main/timm/layers/drop.py#L137
    """

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        self.drop_prob, self.scale_by_keep = drop_prob, scale_by_keep
        super().__init__()

    def forward(self, inputs):
        if self.drop_prob == 0.0 or not self.training:
            return inputs
        keep_prob = 1 - self.drop_prob
        shape = (inputs.shape[0],) + (1,) * (inputs.ndim - 1)  # work with diff dim tensors, not just 2D ConvNets
        random_tensor = inputs.new_empty(shape).bernoulli_(keep_prob)
        if keep_prob > 0.0 and self.scale_by_keep:
            random_tensor.div_(keep_prob)
        return inputs * random_tensor

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob,3):0.3f}"


class Dropout(Layer):
    def __init__(self, rate, noise_shape=None, **kwargs):
        self.rate, self.noise_shape = rate, noise_shape
        super().__init__(**kwargs)

    def build(self, input_shape):
        if self.noise_shape is None:
            dims, rate = len(input_shape), self.rate
            # self.module = torch.nn.Dropout1d(p=rate) if dims == 3 else (torch.nn.Dropout2d(p=rate) if dims == 4 else torch.nn.Dropout3d(p=rate))
            self.module = torch.nn.Dropout(p=rate)
        else:
            self.module = _DropPath(drop_prob=self.rate)
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"rate": self.rate, "noise_shape": self.noise_shape})
        return config


class Flatten(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = nn.Flatten()
        super().build(input_shape)

    def compute_output_shape(self, inptut_shape):
        return [None, int(np.prod(inptut_shape[1:]))]


class _Pooling2D(_BaseConvPool):
    def __init__(self, pool_size=(2, 2), strides=1, padding="valid", reduce="mean", **kwargs):
        super().__init__(kernel_size=pool_size, strides=strides, padding=padding, dilation_rate=(1, 1), **kwargs)
        self.pool_size = pool_size if isinstance(pool_size, (list, tuple)) else [pool_size, pool_size]
        self.reduce = reduce

    def build_module(self, input_shape):
        if self.reduce.lower() == "max":
            module = nn.MaxPool2d(kernel_size=self.pool_size, stride=self.strides, padding=self._pad)
        else:
            module = nn.AvgPool2d(kernel_size=self.pool_size, stride=self.strides, padding=self._pad, count_include_pad=False)
        return module

    def compute_output_shape(self, input_shape):
        output_size = compute_conv_output_size(input_shape, self.pool_size, self.strides, self.padding, dilation_rate=1)
        return [None, input_shape[1], *output_size]

    def get_config(self):
        config = super().get_config()
        config.pop("kernel_size")  # From super
        config.pop("dilation_rate")  # From super
        config.update({"pool_size": self.pool_size, "strides": self.strides, "padding": self.padding})  # Not saving reduce
        return config


class AvgPool2D(_Pooling2D):
    def __init__(self, pool_size=(2, 2), strides=1, padding="valid", **kwargs):
        super().__init__(pool_size=pool_size, strides=strides, padding=padding, reduce=kwargs.pop("reduce", "mean"), **kwargs)


class MaxPool2D(_Pooling2D):
    def __init__(self, pool_size=(2, 2), strides=None, padding="valid", **kwargs):
        strides = strides or pool_size
        super().__init__(pool_size=pool_size, strides=strides, padding=padding, reduce=kwargs.pop("reduce", "max"), padding_value=-np.inf, **kwargs)


class GlobalAveragePooling1D(Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = torch.nn.Sequential(torch.nn.AdaptiveAvgPool1d(1), torch.nn.Flatten(1))
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape[:2]


class GlobalAveragePooling2D(Layer):
    def __init__(self, keepdims=False, **kwargs):
        self.keepdims = keepdims
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = torch.nn.AdaptiveAvgPool2d(1) if self.keepdims else torch.nn.Sequential(torch.nn.AdaptiveAvgPool2d(1), torch.nn.Flatten(1))
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        return (input_shape[:2] + [1] * (len(input_shape) - 2)) if self.keepdims else input_shape[:2]

    def get_config(self):
        config = super().get_config()
        config.update({"keepdims": self.keepdims})
        return config


class LeakyReLU(Layer):
    def __init__(self, alpha=0.3, **kwargs):
        self.alpha = alpha
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = nn.LeakyReLU(negative_slope=self.alpha)
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"alpha": self.alpha})
        return config


class Permute(Layer):
    def __init__(self, dims, **kwargs):
        self.dims = dims
        super().__init__(**kwargs)
        assert sorted(dims) == list(range(1, len(dims) + 1)), "The set of indices in `dims` must be consecutive and start from 1. dims: {}".format(dims)

    def build(self, input_shape):
        self.module = partial(torch.permute, dims=[0, *self.dims])
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        # output_shape = input_shape.copy()
        # for i, dim in enumerate(self.dims):
        #     output_shape[i + 1] = input_shape[dim]
        return [input_shape[0]] + [input_shape[dim] for dim in self.dims]

    def get_config(self):
        config = super().get_config()
        config.update({"dims": self.dims})
        return config

    def extra_repr(self):
        return f"dims={self.dims}"


class _Reshape(nn.Module):
    def __init__(self, target_shape, **kwargs):
        self.target_shape = [-1 if ii is None else ii for ii in target_shape]
        super().__init__(**kwargs)

    def forward(self, inputs):
        return inputs.contiguous().view([*self.target_shape])

    def extra_repr(self):
        return f"target_shape={self.target_shape}"


class _ReshapeDynamic(Layer):
    """
    Reshape supporting dynamic size.
    >>> from keras_cv_attention_models.pytorch_backend import layers, functional, models
    >>> inputs = layers.Input([3, None, None])
    >>> nn = layers.Reshape([3, -1])(inputs)
    >>> nn = layers._ReshapeDynamic(inputs.shape)([nn, functional.shape(inputs)])
    >>> mm = models.Model(inputs, nn)
    >>> print(f"{mm.input_shape = }, {mm.output_shape = }")
    >>> # mm.input_shape = (None, 3, None, None), mm.output_shape = (None, 3, None, None)
    >>> print(f"{mm(torch.ones([1, 3, 24, 24])).shape = }")
    >>> # mm(torch.ones([1, 3, 24, 24])).shape = torch.Size([1, 3, 24, 24])
    >>> print(f"{mm(torch.ones([1, 3, 24, 42])).shape = }")
    >>> # mm(torch.ones([1, 3, 24, 42])).shape = torch.Size([1, 3, 24, 42])
    """

    def __init__(self, target_shape, **kwargs):
        self.target_shape = target_shape  # Required for compute_output_shape, as it's not able to get
        super().__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 2, "Input should be in format [inputs, target_shape]"
        self.module = lambda inputs: inputs[0].contiguous().view(list(inputs[1]))
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        return [None if ii == -1 else ii for ii in self.target_shape]

    def get_config(self):
        config = super().get_config()
        config.update({"target_shape": self.target_shape})
        return config


class Reshape(Layer):
    def __init__(self, target_shape, **kwargs):
        self.target_shape = [-1 if ii is None else ii for ii in target_shape]
        super().__init__(**kwargs)

    def build(self, input_shape):
        num_unknown_dim = sum([ii == -1 for ii in self.target_shape])
        assert num_unknown_dim < 2, "At most one unknown dimension in output_shape: {}".format(self.target_shape)

        if all([ii is not None and ii != -1 for ii in input_shape[1:]]) and num_unknown_dim == 0:  # Static input_shape
            total_size = np.prod(input_shape[1:])
            assert total_size == np.prod(self.target_shape), "Total size of new array must be unchanged, {} -> {}".format(input_shape, self.target_shape)
        self.module = partial(lambda inputs: inputs.contiguous().view([inputs.shape[0], *self.target_shape]))
        #     if num_unknown_dim > 0:
        #         # known_target_dim = -1 * np.prod(self.target_shape)
        #         self.module = partial(lambda inputs: inputs.contiguous().view([inputs.shape[0]] + self.target_shape))
        #                 # [inputs.shape[0]] + [inputs.shape[1:].numel() // known_target_dim if ii == -1 else ii for ii in self.target_shape]))
        #     else:
        #         self.module = partial(lambda inputs: inputs.contiguous().view([inputs.shape[0], *self.target_shape]))
        # else:
        #     total_size = np.prod(input_shape[1:])
        #     if num_unknown_dim > 0:
        #         unknown_dim = total_size // (-1 * np.prod(self.target_shape))
        #         target_shape = [unknown_dim if ii == -1 else ii for ii in self.target_shape]
        #         self.module = partial(lambda inputs: inputs.contiguous().view([inputs.shape[0], *self.target_shape]))
        #     else:
        #         target_shape = self.target_shape
        #         self.module = partial(lambda inputs: inputs.contiguous().view([-1, *self.target_shape]))
        #     assert total_size == np.prod(target_shape), "Total size of new array must be unchanged, {} -> {}".format(input_shape, self.target_shape)

        # self.module = partial(torch.reshape, shape=[-1, *self.target_shape])
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        num_unknown_dim = sum([ii == -1 for ii in self.target_shape])
        if all([ii is not None and ii != -1 for ii in input_shape[1:]]) and num_unknown_dim == 1:  # Static input_shape
            total_size = np.prod(input_shape[1:])
            unknown_dim = total_size // (-1 * np.prod(self.target_shape))
            target_shape = [unknown_dim if ii == -1 else ii for ii in self.target_shape]
        else:
            target_shape = self.target_shape
        return [input_shape[0]] + [None if ii == -1 else ii for ii in target_shape]

    def get_config(self):
        config = super().get_config()
        config.update({"target_shape": self.target_shape})
        return config

    def extra_repr(self):
        return f"target_shape={self.target_shape}"


class _ResizeDynamic(Layer):
    """
    Resize supporting dynamic size.
    >>> import torch
    >>> from keras_cv_attention_models.pytorch_backend import layers
    >>> inputs = layers.Input([3, 16, 16])
    >>> size = layers.Shape(inputs)[2:]
    >>> aa = layers._ResizeDynamic()
    >>> print(f"{aa([torch.ones([1, 3, 24, 24]), np.array(size)]).shape = }")
    >>> # aa(torch.ones([1, 3, 24, 24])).shape = torch.Size([1, 3, 16, 16])
    >>> inputs.set_shape([None, 1, 32, 32])
    >>> print(f"{aa([torch.ones([1, 3, 24, 24]), np.array(size)]).shape = }")
    >>> # aa(torch.ones([1, 3, 24, 24])).shape = torch.Size([1, 3, 32, 32])
    """

    def __init__(self, method="bilinear", preserve_aspect_ratio=False, antialias=False, **kwargs):
        self.method, self.preserve_aspect_ratio, self.antialias = method, preserve_aspect_ratio, antialias
        super().__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) == 2, "Input should be in format [inputs, size]"
        input_shape, size_shape = input_shape[0], input_shape[1]

        size_len = size_shape[0] if isinstance(size_shape, (list, tuple)) else 1
        assert len(input_shape) - 2 == size_len, f"Provided input_shape={input_shape} length should be larger than size_shape={size_shape} by 2"
        if self.antialias:
            assert len(input_shape) == 4, "Anti-alias option is only supported for 2D resize"
            assert self.method in ["bilinear", "bicubic"], "Anti-alias option is only supported for bilinear and bicubic modes"

        self.module = lambda inputs: torch.functional.F.interpolate(inputs[0], size=list(inputs[1]), mode=self.method, antialias=self.antialias)
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        # input_shape = list(input_shape[0])
        return list(input_shape[0][:2]) + [None] * (len(input_shape[0]) - 2)  # Dynamic shape

    def get_config(self):
        config = super().get_config()
        config.update({"method": self.method, "preserve_aspect_ratio": self.preserve_aspect_ratio, "antialias": self.antialias})
        return config


class Softmax(Layer):
    def __init__(self, axis=1, **kwargs):
        self.axis = axis
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = partial(torch.softmax, dim=self.axis)
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"axis": self.axis})
        return config


class UpSampling2D(Layer):
    def __init__(self, size=(2, 2), data_format=None, interpolation="nearest", **kwargs):
        self.data_format, self.interpolation = data_format, interpolation
        self.size = tuple(size if isinstance(size, (list, tuple)) else [size, size])
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.module = nn.Upsample(scale_factor=self.size, mode=self.interpolation)
        super().build(input_shape)

    def get_config(self):
        config = super().get_config()
        config.update({"size": self.size, "data_format": self.data_format, "interpolation": self.interpolation})
        return config

    def compute_output_shape(self, input_shape):
        height = None if input_shape[2] is None else input_shape[2] * self.size[0]
        width = None if input_shape[3] is None else input_shape[3] * self.size[1]
        return [None, input_shape[1], height, width]


class ZeroPadding2D(Layer):
    def __init__(self, padding=(1, 1), **kwargs):
        assert len(padding) == 2 if isinstance(padding, (list, tuple)) else isinstance(padding, int), "padding should be 2 values or an int: {}".format(padding)
        if isinstance(padding, (list, tuple)) and isinstance(padding[0], (list, tuple)):
            assert isinstance(padding[1], (list, tuple)) and len(padding[0]) == 2 and len(padding[1]) == 2

        self.padding = list(padding) if isinstance(padding, (list, tuple)) else [padding, padding]
        if isinstance(padding, (list, tuple)):
            if isinstance(padding[0], (list, tuple)):
                self.padding = padding
            else:
                self.padding = [[padding[0], padding[0]], [padding[1], padding[1]]]
        else:
            self.padding = [[padding, padding], [padding, padding]]
        super().__init__(**kwargs)

    def build(self, input_shape):
        padding = [self.padding[1][0], self.padding[1][1], self.padding[0][0], self.padding[0][1]]  # [left, right, top, bottom]
        self.module = torch.nn.ZeroPad2d(padding=padding)
        super().build(input_shape)

    def compute_output_shape(self, input_shape):
        hh = None if input_shape[2] is None else (input_shape[2] + self.padding[0][0] + self.padding[0][1])
        ww = None if input_shape[3] is None else (input_shape[3] + self.padding[1][0] + self.padding[1][1])
        return [input_shape[0], input_shape[1], hh, ww]

    def get_config(self):
        config = super().get_config()
        config.update({"padding": self.padding})
        return config



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    add_with_layer_scale_and_drop_block,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    layer_norm,
    mlp_block,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "hornet_base": {"imagenet": {224: "b5e808fe5996a8cc980552994f5ca875"}},
    "hornet_base_gf": {"imagenet": {224: "c08c8dc1666c4afbc98e856fa7e53c99"}},
    "hornet_large": {"imagenet22k": {224: "27e7a8f18657c82cf9ad07dc21df9a35"}},
    "hornet_large_gf": {"imagenet22k": {224: "ffd3e68dae2a365d0ecb5df2f0ea881b", 384: "1e3bd0f4b63f65ff7aa3397129ff7a23"}},
    "hornet_small_gf": {"imagenet": {224: "79ae2409e29b125ba9978227f353cd7e"}},
    "hornet_small": {"imagenet": {224: "4ac8b8552a750303d65cae5fa0ff8013"}},
    "hornet_tiny_gf": {"imagenet": {224: "49323a163df84196e9d40a65289e54a9"}},
    "hornet_tiny": {"imagenet": {224: "7bdafa31599cb863b42d960b9d7f5ae4"}},
}


@backend.register_keras_serializable(package="kecam/hornet")
class ComplexDense(layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        _, input_height, input_width, channel = input_shape

        param_shape = (2, input_height, input_width, channel)  # 2 means `real, img` for converting to complex
        initializer = initializers.RandomNormal(stddev=0.02)
        self.complex_weight = self.add_weight(name="complex_weight", shape=param_shape, initializer=initializer, trainable=True)
        self.input_height, self.input_width = input_height, input_width

    def call(self, inputs):
        complex_weight = functional.complex(self.complex_weight[0], self.complex_weight[1])
        complex_weight = functional.cast(complex_weight, inputs.dtype)
        return inputs * complex_weight

    def load_resized_weights(self, source_layer, method="bilinear"):
        if isinstance(source_layer, dict):
            source_tt = source_layer["complex_weight:0"]  # weights
        else:
            source_tt = source_layer.complex_weight  # layer
        tt = functional.resize(source_tt, (self.input_height, self.input_width), method=method, antialias=True)
        self.complex_weight.assign(tt)


def global_local_filter(inputs, name=None):
    _, height, width, channel = inputs.shape
    nn = layer_norm(inputs, name=name and name + "pre_")
    dw, fft = functional.split(nn, 2, axis=-1)
    dw = depthwise_conv2d_no_bias(dw, 3, padding="same", use_bias=False, name=name)

    # fft = tf.py_function(lambda xx: np.fft.rfft2(xx, axes=(1, 2), norm='ortho'), [fft], Tout=tf.complex128)
    # np.fft.rfft2(aa, axes=[1, 2]) ==> tf.transpose(tf.signal.rfft2d(tf.transpose(aa, [0, 3, 1, 2])), [0, 2, 3, 1])
    # ortho_norm = float(tf.sqrt(float(height * width)))
    fft = functional.transpose(fft, [0, 3, 1, 2])
    fft = layers.Lambda(functional.rfft2d)(fft)
    fft = functional.transpose(fft, [0, 2, 3, 1])
    # fft /= ortho_norm  # Means `norm='ortho'`, but will multiply back for `irfft2d` anyway, not affecting results.
    # fft.set_shape([None, height, (width + 2) // 2, channel // 2])
    # print(f">>>> {inputs.shape = }, {fft.shape = }, {fft.dtype = }")
    fft = ComplexDense(name=name and name + "complex_dense")(fft)
    # fft = tf.py_function(lambda xx: np.fft.irfft2(xx, s=[height, width], axes=(1, 2), norm='ortho'), [fft], Tout=inputs.dtype)
    # np.fft.irfft2(bb, s=[13, 14], axes=(1, 2)) ==> tf.transpose(tf.signal.irfft2d(tf.transpose(bb, [0, 3, 1, 2]), fft_length=[13, 14]), [0, 2, 3, 1])
    fft = functional.transpose(fft, [0, 3, 1, 2])
    fft = layers.Lambda(lambda xx: functional.irfft2d(xx, fft_length=[height, width]))(fft)
    fft = functional.transpose(fft, [0, 2, 3, 1])
    # fft *= ortho_norm
    # fft = tf.cast(fft, inputs.dtype)
    # fft.set_shape([None, height, width, channel // 2])

    out = functional.concat([functional.expand_dims(dw, -1), functional.expand_dims(fft, -1)], axis=-1)
    out = functional.reshape(out, [-1, height, width, channel])
    out = layer_norm(out, name=name and name + "post_")
    return out


def gnconv(inputs, use_global_local_filter=False, dw_kernel_size=7, gn_split=3, scale=0.3333333, name=None):
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    nn = conv2d_no_bias(inputs, input_channel * 2, kernel_size=1, use_bias=True, name=name and name + "pre_")
    split_dims = [input_channel // (2**ii) for ii in range(gn_split)][::-1]
    # print(f">>>> {nn.shape = }, {split_dims = }")
    pw_first, dw_list = functional.split(nn, [split_dims[0], sum(split_dims)], axis=channel_axis)

    if use_global_local_filter:
        dw_list = global_local_filter(dw_list, name=name and name + "gf_")
    else:
        dw_list = depthwise_conv2d_no_bias(dw_list, kernel_size=dw_kernel_size, padding="same", use_bias=True, name=name and name + "list_")
    dw_list *= scale

    dw_list = functional.split(dw_list, split_dims, axis=channel_axis)
    nn = pw_first * dw_list[0]
    for id, dw in enumerate(dw_list[1:], start=1):
        pw = conv2d_no_bias(nn, dw.shape[channel_axis], kernel_size=1, use_bias=True, name=name and name + "pw{}_".format(id))
        nn = pw * dw

    nn = conv2d_no_bias(nn, input_channel, kernel_size=1, use_bias=True, name=name and name + "output_")
    return nn


def block(inputs, mlp_ratio=4, use_global_local_filter=False, gn_split=3, scale=0.3333333, layer_scale=0, drop_rate=0, activation="gelu", name=""):
    # print(global_query)
    attn = layer_norm(inputs, name=name + "attn_")
    attn = gnconv(attn, use_global_local_filter, gn_split=gn_split, scale=scale, name=name + "gnconv_")
    attn_out = add_with_layer_scale_and_drop_block(inputs, attn, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "1_")

    mlp = attn_out if backend.image_data_format() == "channels_last" else layers.Permute((2, 3, 1))(attn_out)
    mlp = layer_norm(mlp, axis=-1, name=name + "mlp_")
    mlp = mlp_block(mlp, int(mlp.shape[-1] * mlp_ratio), use_conv=False, activation=activation, name=name + "mlp_")
    mlp = mlp if backend.image_data_format() == "channels_last" else layers.Permute((3, 1, 2))(mlp)
    return add_with_layer_scale_and_drop_block(attn_out, mlp, layer_scale=layer_scale, drop_rate=drop_rate, name=name + "2_")


def HorNet(
    num_blocks=[2, 3, 18, 2],
    embed_dim=64,
    mlp_ratio=4,
    gn_split=[2, 3, 4, 5],
    use_global_local_filter=False,
    scale=0.3333333,
    layer_scale=1e-6,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="hornet",
    kwargs=None,
):
    """Patch stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = conv2d_no_bias(inputs, embed_dim, kernel_size=4, strides=4, use_bias=True, name="stem_")
    nn = layer_norm(nn, name="stem_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, num_block in enumerate(num_blocks):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = layer_norm(nn, name=stack_name)
            input_channels = nn.shape[-1 if backend.image_data_format() == "channels_last" else 1]
            nn = conv2d_no_bias(nn, input_channels * 2, kernel_size=2, strides=2, use_bias=True, name=stack_name)

        cur_use_global_local_filter = use_global_local_filter[stack_id] if isinstance(use_global_local_filter, (list, tuple)) else use_global_local_filter
        cur_gn_split = gn_split[stack_id] if isinstance(gn_split, (list, tuple)) else gn_split
        cur_scale = scale[stack_id] if isinstance(scale, (list, tuple)) else scale

        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = block(nn, mlp_ratio, cur_use_global_local_filter, cur_gn_split, cur_scale, layer_scale, block_drop_rate, activation, name=block_name)
            global_block_id += 1

    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool")(nn)
        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        nn = layer_norm(nn, name="pre_output_")
        nn = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="predictions")(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "hornet", pretrained, mismatch_class=ComplexDense)
    return model


@register_model
def HorNetTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return HorNet(**locals(), model_name="hornet_tiny", **kwargs)


@register_model
def HorNetTinyGF(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    use_global_local_filter = [False, False, True, True]
    return HorNet(**locals(), model_name="hornet_tiny_gf", **kwargs)


@register_model
def HorNetSmall(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 96
    return HorNet(**locals(), model_name="hornet_small", **kwargs)


@register_model
def HorNetSmallGF(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 96
    use_global_local_filter = [False, False, True, True]
    return HorNet(**locals(), model_name="hornet_small_gf", **kwargs)


@register_model
def HorNetBase(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 128
    return HorNet(**locals(), model_name="hornet_base", **kwargs)


@register_model
def HorNetBaseGF(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 128
    use_global_local_filter = [False, False, True, True]
    return HorNet(**locals(), model_name="hornet_base_gf", **kwargs)


@register_model
def HorNetLarge(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet22k", **kwargs):
    embed_dim = 192
    return HorNet(**locals(), model_name="hornet_large", **kwargs)


@register_model
def HorNetLargeGF(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet22k", **kwargs):
    embed_dim = 192
    use_global_local_filter = [False, False, True, True]
    return HorNet(**locals(), model_name="hornet_large_gf", **kwargs)



import math
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    add_pre_post_process,
)
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models.download_and_load import reload_model_weights
from keras_cv_attention_models.coco import eval_func, anchors_func

PRETRAINED_DICT = {
    "yolox_nano": {"coco": "7c97d60d4cc9d54321176f844acee627"},
    "yolox_tiny": {"coco": "f9b51ff24290090c86a10a45f811140b"},
    "yolox_s": {"coco": "a989f5a808ddc4a8242157a6a3e64977"},
    "yolox_m": {"coco": "5c2333d2f12b2f48e3ec8555b29d242f"},
    "yolox_l": {"coco": "a07c48994b7a67dba421025ef39b858b"},
    "yolox_x": {"coco": "de9741d3f67f50c54856bcae0f07b7ef"},
}


""" CSPDarknet backbone """
BATCH_NORM_EPSILON = 1e-3
BATCH_NORM_MOMENTUM = 0.97


def conv_dw_pw_block(inputs, filters, kernel_size=1, strides=1, use_depthwise_conv=False, activation="swish", name=""):
    nn = inputs
    if use_depthwise_conv:
        nn = depthwise_conv2d_no_bias(nn, kernel_size, strides, padding="same", name=name)
        nn = batchnorm_with_activation(nn, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name + "dw_")
        kernel_size, strides = 1, 1
    nn = conv2d_no_bias(nn, filters, kernel_size, strides, padding="same", name=name)
    nn = batchnorm_with_activation(nn, activation=activation, epsilon=BATCH_NORM_EPSILON, momentum=BATCH_NORM_MOMENTUM, name=name)
    return nn


def csp_block(inputs, expansion=0.5, use_shortcut=True, use_depthwise_conv=False, activation="swish", name=""):
    input_channels = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    nn = conv_dw_pw_block(inputs, int(input_channels * expansion), activation=activation, name=name + "1_")
    nn = conv_dw_pw_block(nn, input_channels, kernel_size=3, strides=1, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "2_")
    if use_shortcut:
        nn = layers.Add()([inputs, nn])
    return nn


def csp_stack(inputs, depth, out_channels=-1, expansion=0.5, use_shortcut=True, use_depthwise_conv=False, activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    out_channels = inputs.shape[channel_axis] if out_channels == -1 else out_channels
    hidden_channels = int(out_channels * expansion)
    short = conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "short_")

    deep = conv_dw_pw_block(inputs, hidden_channels, kernel_size=1, activation=activation, name=name + "deep_")
    for id in range(depth):
        block_name = name + "block{}_".format(id + 1)
        deep = csp_block(deep, 1, use_shortcut=use_shortcut, use_depthwise_conv=use_depthwise_conv, activation=activation, name=block_name)

    out = functional.concat([deep, short], axis=channel_axis)
    out = conv_dw_pw_block(out, out_channels, kernel_size=1, activation=activation, name=name + "output_")
    return out


def spatial_pyramid_pooling(inputs, pool_sizes=(5, 9, 13), activation="swish", name=""):
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    input_channels = inputs.shape[channel_axis]
    nn = conv_dw_pw_block(inputs, input_channels // 2, kernel_size=1, activation=activation, name=name + "1_")
    pp = [layers.MaxPool2D(pool_size=ii, strides=1, padding="same")(nn) for ii in pool_sizes]
    nn = functional.concat([nn, *pp], axis=channel_axis)
    nn = conv_dw_pw_block(nn, input_channels, kernel_size=1, activation=activation, name=name + "2_")
    return nn


def focus_stem(inputs, filters, kernel_size=3, strides=1, padding="valid", activation="swish", name=""):
    is_channels_last = image_data_format() == "channels_last"
    channel_axis = -1 if is_channels_last else 1
    if padding.lower() == "same":  # Handling odd input_shape
        inputs = functional.pad(inputs, [[0, 0], [0, 1], [0, 1], [0, 0]] if is_channels_last else [[0, 0], [0, 0], [0, 1], [0, 1]])
        patch_top_left = inputs[:, :-1:2, :-1:2] if is_channels_last else inputs[:, :, :-1:2, :-1:2]
        patch_top_right = inputs[:, :-1:2, 1::2] if is_channels_last else inputs[:, :, :-1:2, 1::2]
        patch_bottom_left = inputs[:, 1::2, :-1:2] if image_data_format() == "channels_last" else inputs[:, :, 1::2, :-1:2]
        patch_bottom_right = inputs[:, 1::2, 1::2] if image_data_format() == "channels_last" else inputs[:, :, 1::2, 1::2]
    else:
        patch_top_left = inputs[:, ::2, ::2] if is_channels_last else inputs[:, :, ::2, ::2]
        patch_top_right = inputs[:, ::2, 1::2] if is_channels_last else inputs[:, :, ::2, 1::2]
        patch_bottom_left = inputs[:, 1::2, ::2] if is_channels_last else inputs[:, :, 1::2, ::2]
        patch_bottom_right = inputs[:, 1::2, 1::2] if is_channels_last else inputs[:, :, 1::2, 1::2]
    nn = functional.concat([patch_top_left, patch_bottom_left, patch_top_right, patch_bottom_right], axis=channel_axis)
    nn = conv_dw_pw_block(nn, filters, kernel_size=kernel_size, strides=strides, activation=activation, name=name)
    return nn


def CSPDarknet(width_mul=1, depth_mul=1, out_features=[-3, -2, -1], use_depthwise_conv=False, input_shape=(512, 512, 3), activation="swish", model_name=""):
    base_channels, base_depth = int(width_mul * 64), max(round(depth_mul * 3), 1)
    inputs = layers.Input(backend.align_input_shape_by_image_data_format(input_shape))

    """ Stem """
    nn = focus_stem(inputs, base_channels, activation=activation, name="stem_")
    features = [nn]

    """ dark blocks """
    depthes = [base_depth, base_depth * 3, base_depth * 3, base_depth]
    channels = [base_channels * 2, base_channels * 4, base_channels * 8, base_channels * 16]
    use_spps = [False, False, False, True]
    use_shortcuts = [True, True, True, False]
    for id, (channel, depth, use_spp, use_shortcut) in enumerate(zip(channels, depthes, use_spps, use_shortcuts)):
        stack_name = "stack{}_".format(id + 1)
        nn = conv_dw_pw_block(nn, channel, kernel_size=3, strides=2, use_depthwise_conv=use_depthwise_conv, activation=activation, name=stack_name)
        if use_spp:
            nn = spatial_pyramid_pooling(nn, activation=activation, name=stack_name + "spp_")
        # nn = SPPBottleneck(base_channels * 16, base_channels * 16, activation=act)
        nn = csp_stack(nn, depth, use_shortcut=use_shortcut, use_depthwise_conv=use_depthwise_conv, activation=activation, name=stack_name)
        features.append(nn)

    nn = [features[ii] for ii in out_features]
    model = models.Model(inputs, nn, name=model_name)
    return model


""" path aggregation fpn """


def upsample_merge(inputs, csp_depth, use_depthwise_conv=False, activation="swish", name=""):
    # print(f">>>> upsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    target_channel = inputs[-1].shape[channel_axis]
    fpn_out = conv_dw_pw_block(inputs[0], target_channel, activation=activation, name=name + "fpn_")

    # inputs[0] = layers.UpSampling2D(size=(2, 2), interpolation="nearest", name=name + "up")(fpn_out)
    size = functional.shape(inputs[-1])[1:-1] if image_data_format() == "channels_last" else functional.shape(inputs[-1])[2:]
    inputs[0] = functional.resize(fpn_out, size, method="nearest")
    nn = functional.concat(inputs, axis=channel_axis)
    nn = csp_stack(nn, csp_depth, target_channel, 0.5, False, use_depthwise_conv, activation=activation, name=name)
    return fpn_out, nn


def downsample_merge(inputs, csp_depth, use_depthwise_conv=False, activation="swish", name=""):
    # print(f">>>> downsample_merge inputs: {[ii.shape for ii in inputs] = }")
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    inputs[0] = conv_dw_pw_block(inputs[0], inputs[-1].shape[channel_axis], 3, 2, use_depthwise_conv, activation=activation, name=name + "down_")
    nn = functional.concat(inputs, axis=channel_axis)
    nn = csp_stack(nn, csp_depth, nn.shape[channel_axis], 0.5, False, use_depthwise_conv, activation=activation, name=name)
    return nn


def path_aggregation_fpn(features, depth_mul=1, use_depthwise_conv=False, activation="swish", name=""):
    # p5 -> fpn_out0 -------------+-> pan_out0
    #       v                     ^
    # p4 -> f_out0 -> fpn_out1 ---+-> pan_out1
    #                 v           ^
    # p3 -----------> pan_out2 ---+-> pan_out2
    csp_depth = max(round(depth_mul * 3), 1)
    p3, p4, p5 = features  # p3: [64, 64, 256], p4: [32, 32, 512], p5: [16, 16, 1024]
    # fpn_out0: [16, 16, 512], f_out0: [32, 32, 512]
    fpn_out0, f_out0 = upsample_merge([p5, p4], csp_depth, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "c3p4_")
    # fpn_out1: [32, 32, 256], pan_out2: [64, 64, 256]
    fpn_out1, pan_out2 = upsample_merge([f_out0, p3], csp_depth, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "c3p3_")
    # pan_out1: [32, 32, 512]
    pan_out1 = downsample_merge([pan_out2, fpn_out1], csp_depth, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "c3n3_")
    # pan_out0: [16, 16, 1024]
    pan_out0 = downsample_merge([pan_out1, fpn_out0], csp_depth, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "c3n4_")
    return [pan_out2, pan_out1, pan_out0]


""" YOLOXHead """


def yolox_head_single(
    inputs, out_channels, num_classes=80, regression_len=4, num_anchors=1, use_depthwise_conv=False, use_object_scores=True, activation="swish", name=""
):
    bias_init = initializers.constant(-math.log((1 - 0.01) / 0.01))

    # stem
    stem = conv_dw_pw_block(inputs, out_channels, activation=activation, name=name + "stem_")

    # cls_convs, cls_preds
    cls_nn = conv_dw_pw_block(stem, out_channels, kernel_size=3, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "cls_1_")
    cls_nn = conv_dw_pw_block(cls_nn, out_channels, kernel_size=3, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "cls_2_")
    cls_out = layers.Conv2D(num_classes * num_anchors, kernel_size=1, bias_initializer=bias_init, name=name + "class_out")(cls_nn)
    cls_out = activation_by_name(cls_out, "sigmoid", name=name + "class_out_")
    cls_out = cls_out if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(cls_out)
    cls_out = layers.Reshape([-1, num_classes], name=name + "class_out_reshape")(cls_out)

    # reg_convs, reg_preds
    reg_nn = conv_dw_pw_block(stem, out_channels, kernel_size=3, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "reg_1_")
    reg_nn = conv_dw_pw_block(reg_nn, out_channels, kernel_size=3, use_depthwise_conv=use_depthwise_conv, activation=activation, name=name + "reg_2_")
    reg_out = layers.Conv2D(regression_len * num_anchors, kernel_size=1, name=name + "regression_out")(reg_nn)
    reg_out = reg_out if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(reg_out)
    reg_out = layers.Reshape([-1, regression_len], name=name + "regression_out_reshape")(reg_out)

    # obj_preds
    if use_object_scores:
        obj_out = layers.Conv2D(1 * num_anchors, kernel_size=1, bias_initializer=bias_init, name=name + "object_out")(reg_nn)
        obj_out = activation_by_name(obj_out, "sigmoid", name=name + "object_out_")
        obj_out = obj_out if image_data_format() == "channels_last" or num_anchors == 1 else layers.Permute([2, 3, 1])(obj_out)
        obj_out = layers.Reshape([-1, 1], name=name + "object_out_reshape")(obj_out)
        return functional.concat([reg_out, cls_out, obj_out], axis=-1)
    else:
        return functional.concat([reg_out, cls_out], axis=-1)


def yolox_head(
    inputs, width_mul=1.0, num_classes=80, regression_len=4, num_anchors=1, use_depthwise_conv=False, use_object_scores=True, activation="swish", name=""
):
    out_channel = int(256 * width_mul)
    outputs = []
    for id, input in enumerate(inputs):
        cur_name = name + "{}_".format(id + 1)
        out = yolox_head_single(
            input, out_channel, num_classes, regression_len, num_anchors, use_depthwise_conv, use_object_scores, activation=activation, name=cur_name
        )
        outputs.append(out)
    # outputs = functional.concat([layers.Reshape([-1, ii.shape[-1]])(ii) for ii in outputs], axis=1)
    outputs = functional.concat(outputs, axis=1)
    return outputs


""" YOLOX models """


def YOLOX(
    backbone=None,
    features_pick=[-3, -2, -1],
    depth_mul=1,
    width_mul=-1,  # -1 means: `min([ii.shape[-1] for ii in features]) / 256` for custom backbones.
    use_depthwise_conv=False,
    regression_len=4,  # bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64
    anchors_mode="anchor_free",
    num_anchors="auto",  # "auto" means: anchors_mode=="anchor_free" -> 1, anchors_mode=="yolor" -> 3, else 9
    use_object_scores="auto",  # "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False
    input_shape=(640, 640, 3),
    num_classes=80,
    activation="swish",
    freeze_backbone=False,
    pretrained=None,
    model_name="yolox",
    pyramid_levels_min=3,  # Init anchors for model prediction.
    anchor_scale="auto",  # Init anchors for model prediction. "auto" means 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4
    rescale_mode="raw",  # For decode predictions, raw means input value in range [0, 255].
    kwargs=None,  # Not using, recieving parameter
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)

    if backbone is None:
        width_mul = width_mul if width_mul > 0 else 1
        backbone = CSPDarknet(width_mul, depth_mul, features_pick, use_depthwise_conv, input_shape, activation=activation, model_name="darknet")
        features = backbone.outputs
    else:
        if isinstance(features_pick[0], str):
            features = [backbone.get_layer(layer_name) for layer_name in features_pick]
        else:
            features = model_surgery.get_pyramide_feature_layers(backbone)
            features = [features[id] for id in features_pick]

        feature_names, features = model_surgery.align_pyramide_feature_output_by_image_data_format(features)
        print(">>>> features:", {ii: jj.shape for ii, jj in zip(feature_names, features)})
        width_mul = width_mul if width_mul > 0 else min([ii.shape[-1] for ii in features]) / 256
        print(">>>> width_mul:", width_mul)

    backbone.trainable = False if freeze_backbone else True
    use_object_scores, num_anchors, anchor_scale = anchors_func.get_anchors_mode_parameters(anchors_mode, use_object_scores, num_anchors, anchor_scale)
    inputs = backbone.inputs[0]

    fpn_features = path_aggregation_fpn(features, depth_mul=depth_mul, use_depthwise_conv=use_depthwise_conv, activation=activation, name="pafpn_")
    outputs = yolox_head(
        fpn_features, width_mul, num_classes, regression_len, num_anchors, use_depthwise_conv, use_object_scores, activation=activation, name="head_"
    )
    outputs = layers.Activation("linear", dtype="float32", name="outputs_fp32")(outputs)
    model = models.Model(inputs, outputs, name=model_name)
    reload_model_weights(model, PRETRAINED_DICT, "yolox", pretrained)

    pyramid_levels = [pyramid_levels_min, pyramid_levels_min + len(features_pick) - 1]  # -> [3, 5]
    post_process = eval_func.DecodePredictions(
        backbone.input_shape[1:], pyramid_levels, anchors_mode, use_object_scores, anchor_scale, regression_len=regression_len
    )
    add_pre_post_process(model, rescale_mode=rescale_mode, post_process=post_process)
    return model


@register_model
def YOLOXNano(input_shape=(416, 416, 3), freeze_backbone=False, num_classes=80, backbone=None, activation="swish", pretrained="coco", **kwargs):
    return YOLOX(**locals(), depth_mul=0.33, width_mul=0.25, use_depthwise_conv=True, model_name=kwargs.pop("model_name", "yolox_nano"), **kwargs)


@register_model
def YOLOXTiny(input_shape=(416, 416, 3), freeze_backbone=False, num_classes=80, backbone=None, activation="swish", pretrained="coco", **kwargs):
    return YOLOX(**locals(), depth_mul=0.33, width_mul=0.375, model_name=kwargs.pop("model_name", "yolox_tiny"), **kwargs)


@register_model
def YOLOXS(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, activation="swish", pretrained="coco", **kwargs):
    return YOLOX(**locals(), depth_mul=0.33, width_mul=0.5, model_name=kwargs.pop("model_name", "yolox_s"), **kwargs)


@register_model
def YOLOXM(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, activation="swish", pretrained="coco", **kwargs):
    return YOLOX(**locals(), depth_mul=0.67, width_mul=0.75, model_name=kwargs.pop("model_name", "yolox_m"), **kwargs)


@register_model
def YOLOXL(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, activation="swish", pretrained="coco", **kwargs):
    return YOLOX(**locals(), depth_mul=1.0, width_mul=1.0, model_name=kwargs.pop("model_name", "yolox_l"), **kwargs)


@register_model
def YOLOXX(input_shape=(640, 640, 3), freeze_backbone=False, num_classes=80, backbone=None, activation="swish", pretrained="coco", **kwargs):
    return YOLOX(**locals(), depth_mul=1.33, width_mul=1.25, model_name=kwargs.pop("model_name", "yolox_x"), **kwargs)



from keras_cv_attention_models.hiera.hiera import Hiera, HieraTiny, HieraSmall, HieraBase, HieraBasePlus, HieraLarge, HieraHuge

__head_doc__ = """
Keras implementation of [Github facebookresearch/hiera](https://github.com/facebookresearch/hiera).
Paper [PDF 2306.00989 Hiera: A Hierarchical Vision Transformer without the Bells-and-Whistles](https://arxiv.org/pdf/2306.00989.pdf).
"""

__tail_doc__ = """  strides: list of int indicates strides for each stack. Default `[1, 2, 2, 2]`.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: one of None or "mae_in1k_ft1k".
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

Hiera.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  embed_dim: basic hidden dims, expand * 2 for each stack.
  num_heads: int or list value for num heads in each stack.
  use_window_attentions: boolean or list value, each value in the list can also be a list of boolean.
      Indicates if use window attention in each stack.
      Element value like `[True, False]` means first one is True, others are False.
  mlp_ratio: expand ratio for mlp blocks hidden channel.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model         | Params  | FLOPs   | Input | Top1 Acc |
  | ------------- | ------- | ------- | ----- | -------- |
  | HieraTiny     | 27.91M  | 4.93G   | 224   | 82.8     |
  | HieraSmall    | 35.01M  | 6.44G   | 224   | 83.8     |
  | HieraBase     | 51.52M  | 9.43G   | 224   | 84.5     |
  | HieraBasePlus | 69.90M  | 12.71G  | 224   | 85.2     |
  | HieraLarge    | 213.74M | 40.43G  | 224   | 86.1     |
  | HieraHuge     | 672.78M | 125.03G | 224   | 86.9     |
"""

HieraTiny.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

HieraSmall.__doc__ = HieraTiny.__doc__
HieraBase.__doc__ = HieraTiny.__doc__
HieraBasePlus.__doc__ = HieraTiny.__doc__
HieraLarge.__doc__ = HieraTiny.__doc__
HieraHuge.__doc__ = HieraTiny.__doc__



from keras_cv_attention_models.iformer.iformer import InceptionTransformer, IFormerSmall, IFormerBase, IFormerLarge

__head_doc__ = """
Keras implementation of [Github sail-sg/iFormer](https://github.com/sail-sg/iFormer).
Paper [PDF 2205.12956 Inception Transformer](https://arxiv.org/pdf/2205.12956.pdf).
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be a constant value like `0.2`,
      or a tuple value like `(0, 0.2)` indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  layer_scales: int or list of int, indicates layer scale init value for each stack. Default `[0, 0, 1e-6, 1e-6]`, 0 for not using.
      [Going deeper with Image Transformers](https://arxiv.org/abs/2103.17239).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `None`.
  pretrained: one of `None` (random initialization) or 'imagenet21k-ft1k' (pre-training on ImageNet21k and fine-tuned ImageNet).
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

InceptionTransformer.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  embed_dims: output channels for each stack.
  num_heads: int or list value indicates heads number for `conv_attention_mixer` blocks in each stack.
  num_attn_low_heads: int or list value indicates attention heads number for `attention_low_frequency_mixer` blocks in each stack.
  pool_sizes: int or list value indicates attention blocks key_value downsample rate in each stack.
  mlp_ratios: int or list value indicates expand ratio for mlp blocks hidden channel in each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs  | Input | Top1 Acc |
  | ------------ | ------ | ------ | ----- | -------- |
  | IFormerSmall | 19.9M  | 4.88G  | 224   | 83.4     |
  |              | 20.9M  | 16.29G | 384   | 84.6     |
  | IFormerBase  | 47.9M  | 9.44G  | 224   | 84.6     |
  |              | 48.9M  | 30.86G | 384   | 85.7     |
  | IFormerLarge | 86.6M  | 14.12G | 224   | 84.6     |
  |              | 87.7M  | 45.74G | 384   | 85.8     |
"""

IFormerSmall.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

IFormerBase.__doc__ = IFormerSmall.__doc__
IFormerLarge.__doc__ = IFormerSmall.__doc__



""" Creates an EfficientNet-EdgeTPU model
Ref impl: https://github.com/tensorflow/tpu/tree/master/models/official/efficientnet/edgetpu
"""
import math
from keras_cv_attention_models.efficientnet.efficientnet_v2 import EfficientNetV2
from keras_cv_attention_models.attention_layers import make_divisible
from keras_cv_attention_models.models import register_model


def get_expanded_width_depth(width, depth, fix_head_stem=False):
    out_channels = [ii * width for ii in [24, 32, 48, 96, 144, 192]]
    depthes = [int(math.ceil(ii * depth)) for ii in [1, 2, 4, 5, 4, 2]]
    first_conv_filter = 32 * width
    output_conv_filter = 1280 * width

    out_channels = [out_channels[0], out_channels[0], *out_channels[1:]]
    depthes = [1, depthes[0] - 1, *depthes[1:]]
    return out_channels, depthes, first_conv_filter, output_conv_filter


def EfficientNetEdgeTPU(
    expands=[-1, 4, 8, 8, 8, 8, 8],  # expands[0] = expands[1] * out_channels[0] / first_conv_filter, as timm using expand on out_channel
    out_channels=[24, 24, 32, 48, 96, 144, 192],
    depthes=[1, 0, 2, 4, 5, 4, 2],  # Add an additional block, as timm using expand on out_channel
    strides=[1, 1, 2, 2, 2, 1, 2],
    se_ratios=[0, 0, 0, 0, 0, 0, 0, 0],
    first_conv_filter=32,
    output_conv_filter=1280,
    kernel_sizes=[3, 3, 3, 3, 5, 5, 5],
    use_shortcuts=[False, False, True, True, True, True, True],
    is_fused=[True, True, True, True, False, False, False],
    is_torch_mode=True,
    drop_connect_rate=0.2,
    pretrained="imagenet",
    activation="relu",
    model_name="EfficientNetEdgeTPU",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    expands[0] = make_divisible(out_channels[0], 8) * expands[1] / make_divisible(first_conv_filter, 8)
    return EfficientNetV2(**locals(), **kwargs)


@register_model
def EfficientNetEdgeTPUSmall(input_shape=(224, 224, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.0, 1.0)
    return EfficientNetEdgeTPU(**locals(), model_name="efficientnet_edgetpu-small", **kwargs)


@register_model
def EfficientNetEdgeTPUMedium(input_shape=(240, 240, 3), num_classes=1000, dropout=0.2, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.0, 1.1)
    return EfficientNetEdgeTPU(**locals(), model_name="efficientnet_edgetpu-medium", **kwargs)


@register_model
def EfficientNetEdgeTPULarge(input_shape=(300, 300, 3), num_classes=1000, dropout=0.3, classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels, depthes, first_conv_filter, output_conv_filter = get_expanded_width_depth(1.2, 1.4)
    return EfficientNetEdgeTPU(**locals(), model_name="efficientnet_edgetpu-large", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, initializers, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    add_with_layer_scale_and_drop_block,
    batchnorm_with_activation,
    conv2d_no_bias,
    # HeadInitializer,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

LAYER_NORM_EPSILON = 1e-6
PRETRAINED_DICT = {
    "fasternet_l": {"imagenet": "42de42d7d4405716e575e25aa58e7445"},
    "fasternet_m": {"imagenet": "2de1a10a5aa1092f82cd6f0104b29c6d"},
    "fasternet_s": {"imagenet": "58f97e2986da1b6793ee4466b4421a16"},
    "fasternet_t0": {"imagenet": "c330a6fa902f17993eba5d734c822551"},
    "fasternet_t1": {"imagenet": "1eb9cb6c77542f5f485efc65c75fa780"},
    "fasternet_t2": {"imagenet": "b4edf4df9e261766fb2e17f0bb50651b"},
}


def block(inputs, mlp_ratio=2, partial_conv_ratio=0.25, layer_scale=1e-6, drop_rate=0, activation="gelu", name=""):
    channel_axis = -1 if backend.image_data_format() == "channels_last" else 1
    input_channel = inputs.shape[channel_axis]
    conv_branch_channels = int(input_channel * partial_conv_ratio)

    conv_branch, non_conv_branch = functional.split(inputs, [conv_branch_channels, -1], axis=channel_axis)
    conv_branch = conv2d_no_bias(conv_branch, conv_branch_channels, kernel_size=3, strides=1, padding="same", name=name + "partial")
    nn = functional.concat([conv_branch, non_conv_branch], axis=channel_axis)

    mlp = conv2d_no_bias(nn, int(input_channel * mlp_ratio), kernel_size=1, name=name + "mlp_1_")
    mlp = batchnorm_with_activation(mlp, activation=activation, name=name + "mlp_")
    mlp = conv2d_no_bias(mlp, input_channel, kernel_size=1, name=name + "mlp_2_")
    return add_with_layer_scale_and_drop_block(inputs, mlp, layer_scale=layer_scale, drop_rate=drop_rate, name=name)


def FasterNet(
    num_blocks=[1, 2, 8, 2],
    embed_dim=40,
    patch_size=4,
    mlp_ratio=2,
    partial_conv_ratio=0.25,
    output_conv_filter=1280,
    layer_scale=0,  # > 0 for applying layer_scale, 0 for not using
    # head_init_scale=0.02,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="fasternet",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)

    """ Stem """
    nn = conv2d_no_bias(inputs, embed_dim, kernel_size=patch_size, strides=patch_size, padding="valid", name="stem_")
    nn = batchnorm_with_activation(nn, activation=None, name="stem_")

    """ Blocks """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, num_block in enumerate(num_blocks):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            input_channel = nn.shape[-1] if backend.image_data_format() == "channels_last" else nn.shape[1]
            nn = conv2d_no_bias(nn, input_channel * 2, kernel_size=2, strides=2, name=stack_name + "downsample_")
            nn = batchnorm_with_activation(nn, activation=None, name=stack_name + "downsample_")
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = block(nn, mlp_ratio, partial_conv_ratio, layer_scale, block_drop_rate, activation, name=block_name)
            global_block_id += 1

    """  Output head """
    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(name="avg_pool", keepdims=True)(nn)
        nn = conv2d_no_bias(nn, output_conv_filter, 1, strides=1, padding="valid", name="post_")
        nn = activation_by_name(nn, activation=activation, name="post_")
        nn = layers.Flatten()(nn)

        if dropout > 0:
            nn = layers.Dropout(dropout, name="head_drop")(nn)
        head_init = initializers.TruncatedNormal(stddev=0.02)  # HeadInitializer(scale=head_init_scale)
        nn = layers.Dense(
            num_classes, dtype="float32", activation=classifier_activation, kernel_initializer=head_init, bias_initializer=head_init, name="predictions"
        )(nn)

    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, pretrained_dict=PRETRAINED_DICT, sub_release="fasternet", pretrained=pretrained)
    return model


@register_model
def FasterNetT0(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return FasterNet(**locals(), model_name="fasternet_t0", **kwargs)


@register_model
def FasterNetT1(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 64
    return FasterNet(**locals(), model_name="fasternet_t1", **kwargs)


@register_model
def FasterNetT2(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 96
    return FasterNet(**locals(), model_name="fasternet_t2", **kwargs)


@register_model
def FasterNetS(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 128
    num_blocks = [1, 2, 13, 2]
    return FasterNet(**locals(), model_name="fasternet_s", **kwargs)


@register_model
def FasterNetM(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 144
    num_blocks = [3, 4, 18, 3]
    return FasterNet(**locals(), model_name="fasternet_m", **kwargs)


@register_model
def FasterNetL(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    embed_dim = 192
    num_blocks = [3, 4, 18, 3]
    return FasterNet(**locals(), model_name="fasternet_l", **kwargs)
